"""Camera module.

This module provides a wrapper for :class:`qamlib.Camera` which is augmented with a
high-level control interface for the underlying camera controls.

"""

from dataclasses import dataclass
from typing import cast

import qamlib
from typing_extensions import Self

from quarklib.camera.events import EventManager
from quarklib.controls import DictGroupControl
from quarklib.controls import GroupControl
from quarklib.controls.builders import GroupControlDefinition
from quarklib.controls.qtec_definitions import QTEC_DEFAULT_CONTROLS


@dataclass
class PixelFormat:
    """PixelFormat helper class."""

    fourcc: str
    channels: list[str]
    bits: int

    @classmethod
    def from_qamlib(cls, format: qamlib.PixelFormat) -> Self:
        """Create from a qamlib pixel format."""
        return cls.from_fourcc(format.pixelformat.fourcc)

    @classmethod
    def from_fourcc(cls, fourcc: str) -> Self:  # noqa C901
        """Parse from V4L2 fourcc code.

        The current supported values are: GREY, Y16, QG08, QG16, RGB3, RGB4,
            BGR3, BGR4, GBRG, GRBG, RGGB, BA81.

        :param fourcc: V4L2 fourcc code.

        :raises ValueError: When an invalid `fourcc` code is provided.

        :returns: The pixel format
        """
        match fourcc:
            case "GREY":
                return cls(fourcc, ["Y"], 8)
            case "Y16":
                return cls(fourcc, ["Y"], 16)
            case "QG08":
                return cls(fourcc, ["G"], 8)
            case "QG16":
                return cls(fourcc, ["G"], 16)
            case "RGB3":
                return cls(fourcc, ["R", "G", "B"], 8)
            case "RGB4":
                return cls(fourcc, ["X", "R", "G", "B"], 8)
            case "BGR3":
                return cls(fourcc, ["B", "G", "R"], 8)
            case "BGR4":
                return cls(fourcc, ["X", "B", "G", "R"], 8)
            case "GBRG":
                return cls(fourcc, ["GBRG"], 8)
            case "GRBG":
                return cls(fourcc, ["GRBG"], 8)
            case "RGGB":
                return cls(fourcc, ["RGGB"], 8)
            case "BA81":
                return cls(fourcc, ["BA81"], 8)
            case _:
                raise ValueError("Invalid fourcc code")

    @property
    def n_channels(self):
        """Get the number of channels.

        :returns: The number of channels.

        """
        return len(self.channels)

    def get_permutation_to(self, other: Self):
        """Calculate channel permutation between rgb/bgr formats.

        The function calculates the necessary permutation for the specified conversion.

        :param other: The target pixel format.

        :raises ValueError: If `self` or `other` has an invalid pixel format.

        :returns: List of channel indices describing where the channel in the input
            format is placed in the output format.
        """
        valid = ["RGB4", "BGR4", "RGB3", "BGR3"]
        if self.fourcc not in valid or other.fourcc not in valid:
            raise ValueError(f"Unsupported fourcc. Valid values: {valid}")

        out_map = []
        extra = 0
        for c in self.channels:
            if c in other.channels:
                v = other.channels.index(c)
            else:
                v = extra + other.n_channels
                extra += 1
            out_map.append(v)

        return out_map

    def __str__(self):
        """Return the FOURCC code of the format."""
        return self.fourcc


class Camera(qamlib.Camera):
    """High-level wrapper for :class:`qamlib.Camera`.

    This wrapper provides the same basic interface as :class:`qamlib.Camera` but adds
    a high-level interface to the camera's controls. The controls are predefined by
    default but can be customized. The default control set is specified in
    :attr:`quarklib.camera.controls.definitions.qtec_default_controls_builder`.

    :param path: File path or device number for `qamlib.Camera`. If empty, the
        default device is used.
    :param control_builder: The builder to use for creating the camera controls. By
        default, a predefined builder is used.
    """

    def __init__(
        self,
        path: str | int | None = None,
        control_builder: GroupControlDefinition = QTEC_DEFAULT_CONTROLS,
    ):
        super().__init__(path)

        self._controls: DictGroupControl = cast(
            DictGroupControl, control_builder.build_controls(self, self.__class__)
        )
        self._ev = EventManager()
        self._ev.subscribe_controls_recursive(self._controls)
        self._ev.start()

    @property
    def controls(self) -> GroupControl:
        """High-level controls access point."""
        return self._controls

    def __del__(self):
        """Stop the event manager."""
        self._ev.stop()
