"""Event manager."""

# type: ignore
from typing import Protocol

import qamlib

from quarklib.controls import DictGroupControl
from quarklib.controls import DriverControl


class EventCallback(Protocol):
    """Event callback type definition."""

    # flake8: noqa: D102
    def __call__(self, event: qamlib.BaseEvent):
        pass


class ControlEventCallback(Protocol):
    """Control event callback type definition."""

    # flake8: noqa: D102
    def __call__(self, event: qamlib.ControlEvent):
        pass


class EventManager:
    """Dispatches multiple callbacks depending on event type.

    The class simplifies the use of a single `qamlib.EventDevice` to message
    different objects depending on the event source.

    :param device: Device id or file path.
    """

    def __init__(self, device: int | str | None = None):
        if device is None:
            self._ed = qamlib.EventDevice()
        else:
            self._ed = qamlib.EventDevice(device)
        self._ed.set_callback(self._control_callback)

        self._control_map: dict[int, ControlEventCallback] = {}

    def start(self):
        """Start listening for events."""
        self._ed.start()

    def stop(self):
        """Stop listening for events."""
        self._ed.stop()

    def _callback(self, event: qamlib.BaseEvent):
        self._control_callback(event)

    def _control_callback(self, event: qamlib.ControlEvent):
        callback = self._control_map[event.id]
        callback(event)

    def subscribe_control_callback(
        self, control: DriverControl, callback: ControlEventCallback
    ):
        """Subscribe control events from `control` to `callback`.

        :param control: Control object to register events for.
        :param callback: The function to call when events are received.
        """
        self._control_map[control.id] = callback
        self._ed.subscribe(qamlib.EventType.CTRL, control.id)

    def subscribe_controls_recursive(self, tree: DictGroupControl):
        """Subscribe a group of controls to their own internal callbacks.

        The method only subscribes controls that have the
        `DriverControl.requires_ctrl_events` flag set. Callbacks are defined by
        `DriverControl.control_callback` which, by default, uses the event
        information to update the control's formatter range.

        :param tree: The top-level group control to subscribe.
        """
        for c in tree.children:
            if isinstance(c, DriverControl) and c.requires_event_handler:
                self.subscribe_control_callback(c, c.control_callback)
            elif isinstance(c, DictGroupControl):
                self.subscribe_controls_recursive(c)
