# type: ignore
"""Color Conversion Module.

This module provides an API for performing color channel transformations as supported
by the AFE CCM module.

The module provides functions for creating color correction transformations using
the *ColorCheckerClassic* color reference.
"""

from typing import Protocol

import numpy as np
import numpy.typing as npt

from quarklib.camera import Camera
from quarklib.ccm.detector import ColorChecker
from quarklib.ccm.reference import ColorReference


def _create_design_matrix(color, offset: bool = False):
    matrix = np.zeros((3, 12))
    matrix[0, :3] = color
    matrix[1, 4:7] = color
    matrix[2, 8:11] = color

    # If the offset is not used, simply set the respective matrix values to 0.
    matrix[0, 3] = int(offset)
    matrix[1, 7] = int(offset)
    matrix[2, 11] = int(offset)
    return matrix


class Solver(Protocol):
    """Solver interface.

    Produces a 4x5 color transformation matrix (for use in the CCM module) using two
    color references.
    """

    def __call__(
        self,
        source: ColorReference,
        target: ColorReference,
        offset: bool = False,
        *args,
        **kwargs,
    ) -> npt.NDArray[np.float32]:
        """Calculate solution.

        :param source: The source reference.
        :param target: The target reference.
        :param offset: Whether to use or ignore color offset calculation
            (affine transform).
        :param *args: Additional positional arguments.
        :param **kwargs: Additional keyword arguments.
        :returns:
        """
        ...


def linear_color_correction_rgb(
    source: ColorReference,
    target: ColorReference,
    offset: bool = False,
) -> np.ndarray:
    """Least squares optimization.

    The method finds the minimal distance achievable between transformed points and
    target points. Since the algorithm operates in RGB, it is unlikely that the result
    is also the optimal perceived color distance. The algorithm is therefore best suited
    to machine vision tasks where the RGB color space is used.

    :param source: The source color reference.
    :param target: The target color reference.
    :param offset: Whether to use the offset component of the CCM.
    :returns: The optimized color conversion matrix.
    """
    source_rgb = source.rgb()
    target_rgb = target.rgb()

    design_mat = np.zeros((3 * len(source_rgb), 12), dtype=np.float64)

    for i, s in enumerate(source_rgb):
        design_mat[i * 3 : (i + 1) * 3] = _create_design_matrix(s, offset)

    sol, _, _, _ = np.linalg.lstsq(design_mat, target_rgb.flatten(), rcond=None)

    return sol


def perform_color_correction(
    cam: Camera,
    target: ColorReference,
    solver: Solver = linear_color_correction_rgb,
):
    """Calculate and set the color conversion transformation of a camera.

    :param cam: Target camera.
    :param target: Target color reference.
    :param solver: Solver to use for finding the transformation matrix.

    :raises ValueError: If the pixel format is not RGB.
    """
    control = cam.controls.color_conversion
    control.ccm.reset()
    _, frame = cam.get_frame()

    measured_reference = ColorChecker(frame).reference(frame)
    solution = solver(measured_reference, target, offset=False)

    fourcc = cam.get_format().pixelformat.fourcc

    if fourcc not in ["RGB3", "RGB4"]:
        raise ValueError(
            "Camera pixel format is invalid. Must be RGB with "
            "3/4 channels. Actual format read: ",
            fourcc,
        )

    solution = np.reshape(solution, (3, 4))

    control.ccm[:3, :3] = solution[:, :3]
    control.ccm[:3, 4] = solution[:, 3]


if __name__ == "__main__":
    cam = Camera()

    with cam:
        ref = ColorReference.from_file("official_reference.json")
        perform_color_correction(cam, ref)


__all__ = [
    "Solver",
    "linear_color_correction_rgb",
    "ColorChecker",
    "ColorReference",
    "perform_color_correction",
    "reference",
]
