"""Module that implements color checker detectors."""

from dataclasses import dataclass

import cv2 as cv
import numpy as np

from quarklib.ccm.reference import ColorDefinition
from quarklib.ccm.reference import ColorReference


class ColorChecker:
    """Detector for the ColorChecker reference board.

    :param frame: Image to find the board in.
    :param margin: The proportional margin between each color box. The default value
        should generally be appropriate.
    :param min_area: Minimum box area in pixels. This might have to be modified if the
        board is placed far from the sensor.
    :param box_shape: The shape (dimensions) of the board grid. The default value is
        valid for the ColorCheckerClassic board.
    """

    def __init__(
        self,
        frame: np.ndarray,
        margin: float = 0.024,
        min_area: int = 2000,
        box_shape: tuple[int, int] = (6, 4),
    ):
        self._min_area = min_area

        self._setup_boxes(box_shape, margin)

        binary_image = self._preprocessing(frame)
        accepted_candidates = self._contour_detection(binary_image)
        boxes = _extract_box_information(accepted_candidates)

        self.bounding_box = _create_bounding_box(accepted_candidates)
        self.boxes = self._detect_boxes(boxes)

    def reference(self, frame: np.ndarray) -> ColorReference:
        """Measure the detected board colors and return the reference.

        :param frame: Image to measure the colors from. This may be the same frame as
            used in the constructor but does not have to be, as long as the board or
            camera has not been moved.
        :returns: The measured reference.
        """
        refs = []

        for box in self.boxes:
            rect = box.rect.scale_around_center(0.5)
            p1, p2 = rect.p1_int, rect.p2_int
            region = frame[p1[1] : p2[1], p1[0] : p2[0]]
            rgb = np.mean(region, axis=(0, 1))[:3]

            name = str(box.id)

            refs.append(ColorDefinition(name, rgb, None))

        return ColorReference("color_checker_classic", (6, 4), "camera", refs)

    def visualize(self, frame: np.ndarray, reference: ColorReference):
        """Visualize the detected board and reference.

        This method is useful to visualize the perceived quality of a color-corrected
        image when compared to the ground truth reference.

        :param frame: Image to draw the visualization onto.
        :param reference: Color reference painted onto the image to visualize the color
            difference.
        :returns:
        """
        frame = cv.drawContours(
            frame,
            [self.bounding_box.points],
            contourIdx=0,
            color=(255, 0, 0),
            thickness=2,
            lineType=cv.LINE_AA,
        )

        for box in self.boxes:
            color = (0, 0, 255) if box.imputed else (0, 255, 0)

            frame = cv.rectangle(  # type: ignore
                frame, np.int32(box.rect.p1), np.int32(box.rect.p2), color, thickness=2
            )

            scaled_box = box.rect.scale_around_center(0.5)
            frame = cv.rectangle(  # type: ignore
                frame,
                np.int32(scaled_box.p1),
                np.int32(scaled_box.p2),
                (255, 0, 0),
                thickness=2,
            )

            pos = list(np.int32(box.rect.p1))  # type: ignore

            pos[0] += 10
            pos[1] += 40

            frame = cv.putText(
                frame,
                f"{box.id}:{box.coordinates}",
                org=pos,
                fontFace=cv.FONT_HERSHEY_SIMPLEX,
                fontScale=0.8,
                color=color,
                thickness=2,
                lineType=cv.LINE_AA,
            )
        return frame

    def _setup_boxes(self, box_shape, margin):
        margin_total_x = margin * (box_shape[0] - 1)
        margin_total_y = margin * (box_shape[1] - 1)
        box_w = (1 - margin_total_x) / box_shape[0]
        box_h = (1 - margin_total_y) / box_shape[1]
        self._box_size = (box_w, box_h)
        self._box_shape = box_shape
        self._box_centers = np.zeros((box_shape[1], box_shape[0], 2))
        self._box_centers_flat = np.reshape(
            self._box_centers, (self._box_shape[0] * self._box_shape[1], 2)
        )
        for i in range(box_shape[1]):
            for j in range(box_shape[0]):
                y = i * (box_h + margin) + box_h / 2
                x = j * (box_w + margin) + box_w / 2
                self._box_centers[i, j] = [x, y]

    def _make_box(self, cx, cy, scale, r=4):
        rx = r * scale[0]
        ry = r * scale[1]
        return [
            [int(cx - self._box_size[0] / rx), int(cy - self._box_size[1] / ry)],
            [int(cx + self._box_size[0] / rx), int(cy + self._box_size[1] / ry)],
        ]

    def _detect_boxes(self, boxes):
        final_boxes = []
        missing = set(range(self._box_shape[0] * self._box_shape[1]))
        box_idxs = []
        for box in boxes:
            idx = np.linalg.norm(
                self._box_centers_flat - box.normalize(self.bounding_box).center, axis=1
            ).argmin()
            box_idxs.append(idx)

            missing.remove(idx)
            final_boxes.append(CCBox(box, imputed=False, id=idx))
        for idx in missing:
            cx, cy = self._box_centers_flat[idx]
            p1 = np.array([cx - self._box_size[0] / 2, cy - self._box_size[1] / 2])
            p2 = np.array([cx + self._box_size[0] / 2, cy + self._box_size[1] / 2])

            rect = Rectangle(p1, p2)
            final_boxes.append(
                CCBox(rect.unnormalize(self.bounding_box), imputed=True, id=idx)
            )
            box_idxs.append(idx)

        b = np.array(final_boxes)[np.argsort(box_idxs)]
        return b

    @staticmethod
    def _preprocessing(image):
        image = cv.fastNlMeansDenoising(
            image, h=10, templateWindowSize=7, searchWindowSize=21
        )
        gray = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
        thresh = cv.adaptiveThreshold(
            gray,
            255,
            cv.ADAPTIVE_THRESH_MEAN_C,
            cv.THRESH_BINARY,
            blockSize=21,
            C=3,
        )
        kernel = cv.getStructuringElement(cv.MORPH_RECT, (3, 3))
        thresh = cv.erode(thresh, kernel)
        thresh = cv.dilate(thresh, kernel)
        return thresh

    def _contour_detection(self, thresh):
        contours, hierarchy = cv.findContours(
            thresh, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
        )
        accepted_candidates = []
        for c, h in zip(contours, hierarchy[0], strict=True):
            self._detect_candidate(accepted_candidates, c, h)
        return accepted_candidates

    def _detect_candidate(self, accepted_candidates, contour, hierarchy_info):
        mod = cv.approxPolyDP(
            contour, epsilon=0.1 * cv.arcLength(contour, True), closed=True
        )
        ret = cv.matchShapes(
            mod,
            np.array([[0, 0], [0, 1], [1, 1], [1, 0]]),
            cv.CONTOURS_MATCH_I2,
            parameter=0.0,
        )
        if (
            ret < 0.015
            and cv.contourArea(contour) > self._min_area
            and _is_leaf_node(hierarchy_info)
        ):
            accepted_candidates.append(mod)


@dataclass
class Rectangle:
    """Rectangle utility class.

    :param p1: Lower left point.
    :param p2: Upper right point.
    """

    p1: np.ndarray
    p2: np.ndarray

    @property
    def center(self):
        """The rectangle center point."""
        return (self.p1 + self.p2) / 2

    @property
    def size(self):
        """The rectangle size."""
        return self.p2 - self.p1

    @property
    def p1_int(self):
        """Convert p1 to integer. Useful for drawing."""
        return np.int32(self.p1)

    @property
    def p2_int(self):
        """Convert p1 to integer. Useful for drawing."""
        return np.int32(self.p2)

    @classmethod
    def from_contour(cls, contour: np.ndarray) -> "Rectangle":
        """Create rectangle from contour points."""
        x, y, w, h = cv.boundingRect(contour)
        p1 = np.array([x, y])
        p2 = np.array([x + w, y + h])
        return cls(p1, p2)

    def scale_around_center(self, factor: float) -> "Rectangle":
        """Scale the rectangle around the center point.

        :param factor: Scaling factor. 1 means the rectangle is unchanged.
        :returns: A new scaled rectangle.
        """
        new_size = self.size * factor
        p1 = self.center - new_size / 2
        p2 = self.center + new_size / 2
        return Rectangle(p1, p2)

    def normalize(self, bounding_box: "Rectangle"):
        """Normalize rectangle with respect to a bounding box.

        The resulting rectangle's points will be relative to the bounding box, i.e.,
        point [0, 0] corresponds to the lower left point of `bounding_box` and [1, 1]
        corresponds to the upper right point.

        This method is useful when changing coordinate spaces.

        :param bounding_box: The reference rectangle.
        :returns: The scaled rectangle.
        """
        p1 = bounding_box._normalize_single(self.p1)
        p2 = bounding_box._normalize_single(self.p2)

        return Rectangle(p1, p2)

    def unnormalize(self, bounding_box: "Rectangle"):
        """Reverses the :meth:`Rectangle.normalize` function.

        Use this method to recover the original rectangle points for a normalized
        rectangle.

        :param bounding_box: The reference rectangle (with original points).
        :returns: The scaled rectangle.
        """
        p1 = bounding_box._unnormalize_single(self.p1)
        p2 = bounding_box._unnormalize_single(self.p2)

        return Rectangle(p1, p2)

    @property
    def points(self):
        """Get the four corner points of the rectangle."""
        return np.array(
            [
                [self.p1[0], self.p1[1]],
                [self.p1[0], self.p2[1]],
                [self.p2[0], self.p2[1]],
                [self.p2[0], self.p1[1]],
            ]
        )

    def _normalize_single(self, val: np.ndarray):
        return (val - self.p1) / self.size

    def _unnormalize_single(self, val: np.ndarray):
        return (val * self.size) + self.p1


def _is_leaf_node(hierarchy_info):
    return hierarchy_info[2] == -1


def _create_bounding_box(accepted_candidates) -> Rectangle:
    accepted_candidates = np.array(accepted_candidates)
    _min = accepted_candidates.min(axis=(0, 1, 2))
    _max = accepted_candidates.max(axis=(0, 1, 2))
    return Rectangle(_min, _max)


def _extract_box_information(accepted_candidates) -> list[Rectangle]:
    return [Rectangle.from_contour(c) for c in accepted_candidates]


def _boxes_info_scaled(box_centers, offset, scale):
    return (box_centers - offset) * scale


@dataclass
class CCBox:
    """Helper class that contains information about a detected color box."""

    rect: "Rectangle"
    imputed: bool
    id: int
    columns: int = 6

    @property
    def coordinates(self):
        """Grid coordinates of the box."""
        return np.array([self.id % self.columns, self.id // self.columns])
