"""Generic color reference implementation."""

import json
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt


@dataclass
class ColorReference:
    """Represents a set of reference colors identified by index.

    :param board: An identifier for the physical source reference.
    :param size: Board grid size as (width, height).
    :param source: Reference source location, .e.g. the official reference or
        a combination of image sensor and capture parameters.
    :param colors: The list of references. Each element describes a particular
        color.
    """

    board: str
    size: tuple[int, int]
    source: str

    colors: list["ColorDefinition"]

    @classmethod
    def from_dict(cls, value: dict) -> "ColorReference":
        """Create reference from dictionary.

        The dictionary must contain

        :param value: Dictionary containing as keys the fields specified in
            the class constructor.
        :returns: The reference object.
        """
        return cls(
            value["board"],
            tuple(value["size"]),
            value["source"],
            [ColorDefinition.from_dict(c) for c in value["colors"]],
        )

    @classmethod
    def from_file(cls, path: str) -> "ColorReference":
        """Create reference directly from a file.

        :param path: Path to a valid reference file (generated using
            :meth:`ColorCheckerReference.save`).
        :returns: The reference object.
        """
        with open(path) as file:
            obj = json.load(file)
            return cls.from_dict(obj)

    def save(self, path: str):
        """Save reference to the specified path.

        :param path: JSON file path (must end in the .json extension)
        """
        with open(path, "w") as file:
            json.dump(self.to_dict(), file)

    def rgb(self) -> npt.NDArray[np.float64]:
        """Fetch reference RGB colors as an Nx3 array where N is the number of colors.

        :returns: The array of colors.
        """
        ref_colors = [np.array(ref.rgb, dtype=np.float64) for ref in self.colors]
        return np.stack(ref_colors)

    def lab(self) -> np.ndarray:
        """Fetch reference LAB colors as an Nx3 array where N is the number of colors.

        :raises ValueError: If the color reference does not contain a LAB value.

        :returns: The array of colors.
        """
        ref_colors = [np.array(ref.lab, dtype=np.float64) for ref in self.colors]
        try:
            return np.stack(ref_colors)
        except ValueError:
            raise ValueError(
                "The reference is missing one or more LAB colors."
            ) from ValueError

    def to_dict(self):
        """Convert the object into a dictionary suitable for json serialization.

        :returns: The dictionary.
        """
        return {
            "board": self.board,
            "size": list(self.size),
            "source": self.source,
            "colors": [c.to_dict() for c in self.colors],
        }

    def mse(self, other: "ColorReference") -> float:
        """Mean squared error of each color channel.

        :param other: The comparison reference.
        :returns: A 1x3 array containing the mse of each channel.
        """
        return np.mean(self.se(other), axis=0)

    def se(self, other: "ColorReference") -> npt.NDArray[np.float64]:
        """Squared error of each color patch.

        :param other: The comparison reference.
        :returns: An Nx3 array containing the squared errors of each color.
        """
        return (self.rgb() - other.rgb()) ** 2

    def distance_error(self, other: "ColorReference") -> npt.NDArray[np.float64]:
        """Distance error of each color patch.

        :param other: The comparison reference.
        :returns: An Nx1 array of the Euclidean distance errors of each color.
        """
        return np.sqrt(self.se(other), axis=1)  # type: ignore

    def angular_error(self, other: "ColorReference"):
        """Angular error of each color patch.

        The angular error

        :param other:
        :returns:
        """
        a = self.rgb()
        b = other.rgb()

        return (a @ b.T) / (np.linalg.norm(a, axis=1) * np.linalg.norm(b, axis=1))


@dataclass
class ColorDefinition:
    """Describes a single defined color.

    The definition has a name that may correspond to the official color name.

    :param name: Color name or id.
    :param rgb: Three-element array containing an 8-bit RGB color.
    :param lab: Optional three-element array containing floats describing the color
        in CIELAB space.
    """

    name: str
    rgb: np.ndarray
    lab: np.ndarray | None

    @classmethod
    def from_dict(cls, value: dict):
        """Create definition from dictionary.

        :param value: Dictionary with 'name', 'rgb', and optionally 'lab' keys.
        :returns: The definition object.
        """
        return cls(value["name"], np.array(value["rgb"]), np.array(value.get("lab")))

    def to_dict(self):
        """Convert to dictionary."""
        return {
            "name": self.name,
            "rgb": list(self.rgb),
            "lab": [] if self.lab is None else list(self.lab),
        }
