"""Camera configuration."""

import json
import logging
import re
import typing

import numpy as np
import numpy.typing as npt
import qamlib


log = logging.getLogger("quarklib")
logging.captureWarnings(True)

__PROTECED_CONTROLS = ["adc_gain", "black_level", "vramp"]


def get(  # noqa C901
    cam: qamlib.Camera,
    *,
    controls: list[str] | None = None,
    format: bool = True,
    crop: bool = True,
    framerate: bool = True,
    protect_controls: bool = True,
) -> dict[str, typing.Any]:
    """Get camera configuration.

    :param cam: The camara to save the configuration from.
    :param controls: Names of the controls to save, if ``None`` then all are
      saved.
    :param format: Save the format.
    :param crop: Save the crop(s).
    :param framerate: Save the framerate.
    :param protect_controls: Protect sensor specific controls
      (ADC Gain, Black Level and VRAMP).

    :raises ValueError: If a non-existent control is specified in the configuration.

    :returns: The camera configuration
    """
    ctrls = cam.list_controls()

    if controls is None:
        control_names = ctrls.keys()
    else:
        control_names = list(map(str.lower, controls))

    res: dict[str, typing.Any] = {"version": 0}

    ctrl_values = {}

    for c in control_names:
        if c not in ctrls:
            raise ValueError(f"{c} is not a control for this device")

        ctrl = ctrls[c]
        if (
            ctrl.flags.read_only
            or ctrl.flags.write_only
            or ctrl.flags.disabled
            or ctrl.flags.inactive
        ):
            continue
        if protect_controls and c in __PROTECED_CONTROLS:
            continue

        tmp = cam.get_ext_control(c).to_json()

        if ctrl.type in [qamlib.ControlType.MENU, qamlib.ControlType.INTEGER_MENU]:
            index = cam.get_control(c)
            tmp["menu_item"] = ctrl.items[index]

        ctrl_values[c] = tmp

    res["controls"] = ctrl_values

    if framerate:
        res["framerate"] = cam.get_framerate()

    if format:
        res["format"] = cam.get_format().pixelformat.fourcc

    if crop:
        crops = cam.get_crop()

        save_crops = []
        for c in crops:
            save_crops.append((c.left, c.top, c.width, c.height))

        res["crops"] = save_crops

    return res


def save(
    cam: qamlib.Camera,
    path: str,
    *,
    controls: list[str] | None = None,
    format: bool = True,
    crop: bool = True,
    framerate: bool = True,
    protect_controls: bool = True,
) -> None:
    """Save camera configuration.

    :param cam: The camara to save the configuration from.

    :param path: File path to put the saved configuration.

    :param controls: Names of the controls to save, if ``None`` then all are
      saved.

    :param format: Save the format.

    :param crop: Save the crop(s).

    :param framerate: Save the framerate.

    :param protect_controls: Protect sensor specific controls
      (ADC Gain, Black Level and VRAMP).
    """
    res = get(
        cam,
        controls=controls,
        format=format,
        crop=crop,
        framerate=framerate,
        protect_controls=protect_controls,
    )

    with open(path, "w") as fd:
        fd.write(json.dumps(res))


_REPLACE_REGEX = re.compile(r"[\W]+")
_SIMPLIFY_REGEX = re.compile(r"_+")


# Regex is used for speed instead of doing a for loop over all characters
def _fix_control_name(name: str) -> str:
    name = name.rstrip()
    name = _REPLACE_REGEX.sub("_", name)
    name = _SIMPLIFY_REGEX.sub("_", name)  # Convert "____" to "_"
    return name


def set(  # noqa: C901
    cam: qamlib.Camera,
    config: dict[str, typing.Any],
    *,
    controls: list[str] | None = None,
    format: bool = True,
    crop: bool = True,
    framerate: bool = True,
    protect_controls: bool = True,
):
    """Set camera configuration.

    :param cam: The camera to load the configuration into.
    :param config: The camera configuration to set.
    :param controls: Names of the controls to load, if ``None`` then all are
      saved.
    :param format: Load the format.
    :param crop: Load the crop(s).
    :param framerate: Load the framerate.
    :param protect_controls: Protect sensor specific controls
      (ADC Gain, Black Level and VRAMP).

    :raises ValueError:
        - When setting the crop value, if the configured crop is not a rectangle
            of four points.
        - When specifying `controls`, if `config` does not contain any.
        - When trying to set a control that does not exist on the camera.
        - When trying to set a control that is not included in `config`.
        - When trying to set a control that does not specify a value to set.
        - When using a value with a payload to set a control that does not
            support payloads.
        - When trying to set the trigger sequence with an incomplete value.
        - When trying to set an unsupported control type.
    :raises TypeError: When trying to set an integer control with a non-integer
        value.

    :returns: Nothing.
    """
    if "format" in config and format:
        cam.set_format(config["format"])

    if "crops" in config and crop:
        crops = []
        for rect in config["crops"]:
            if len(rect) != 4:
                raise ValueError(f"A crop rectangle is not of length 4 ({len(rect)}")
            crops.append(qamlib.Rectangle(rect[0], rect[1], rect[2], rect[3]))
        cam.set_crop(crops)

    if "framerate" in config and framerate:
        cam.set_framerate(config["framerate"])

    if "controls" not in config:
        if controls is not None:
            raise ValueError("No controls found in config")
        return

    # Because we updated the control key names in qamlib we have to fix them so
    # if the loaded configuration is old, that it still works.
    new_controls = {}
    for key, value in config["controls"].items():
        new_controls[_fix_control_name(key)] = value
    config["controls"] = new_controls

    if controls is None:
        control_names = config["controls"].keys()
    else:
        control_names = list(map(str.lower, controls))

    cam_ctrls = cam.list_controls()
    res_ctrls = {}
    for c in control_names:
        if protect_controls and c in __PROTECED_CONTROLS:
            continue

        if c not in cam_ctrls:
            raise ValueError(f"{c} not a control for this device")

        if c not in config["controls"]:
            raise ValueError(f"{c} not found in config")

        ctrl = cam_ctrls[c]

        if ctrl.flags.read_only or ctrl.flags.write_only or ctrl.flags.disabled:
            log.info(f"Skipping {c} because it is read only, write only or disabled")
            continue

        if "value" not in config["controls"][c]:
            raise ValueError(f"Control {c} does not have a value")

        value = config["controls"][c]["value"]

        if ctrl.type in [
            qamlib.ControlType.INTEGER,
            qamlib.ControlType.INTEGER64,
            qamlib.ControlType.U8,
            qamlib.ControlType.U16,
            qamlib.ControlType.U32,
        ]:
            if ctrl.flags.has_payload:
                match ctrl.type:
                    case qamlib.ControlType.INTEGER:
                        dtype: npt.DTypeLike = np.int32
                    case qamlib.ControlType.INTEGER64:
                        dtype = np.int64
                    case qamlib.ControlType.U8:
                        dtype = np.uint8
                    case qamlib.ControlType.U16:
                        dtype = np.uint16
                    case _:
                        dtype = np.uint32
                arr = np.array(value, dtype=dtype)
                value = qamlib.ArrayControlValue(arr)
            else:
                if not isinstance(value, int):
                    raise TypeError(f"Control {c} value is not integer type")
                value = qamlib.IntegerControlValue(value)
        elif ctrl.type in [
            qamlib.ControlType.MENU,
            qamlib.ControlType.INTEGER_MENU,
            qamlib.ControlType.BOOLEAN,
            qamlib.ControlType.BITMASK,
        ]:
            if ctrl.flags.has_payload:
                raise ValueError(f"Payload not allowed for control type {ctrl.type}")
            else:
                value = qamlib.IntegerControlValue(value)
        elif ctrl.type == qamlib.ControlType.STRING:
            value = qamlib.StringControlValue(value)
        elif ctrl.type == qamlib.ControlType.TRIG_SEQ:
            value = qamlib.TriggerSequenceValue()
            for val in value:
                if (
                    "exposure_time" not in val
                    or "flash_time" not in val
                    or "frame_delay" not in val
                    or "trigger_delay" not in val
                ):
                    raise ValueError("Incomplete trigger sequence")

                value.add_exposure(
                    val["exposure_time"],
                    val["flash_time"],
                    val["frame_delay"],
                    val["trigger_delay"],
                )
        else:
            raise ValueError(f"Unsupported control type {ctrl.type}")

        res_ctrls[c] = value

    cam.set_controls(res_ctrls)


def load(  # noqa C901
    cam: qamlib.Camera,
    path: str,
    *,
    controls: list[str] | None = None,
    format: bool = True,
    crop: bool = True,
    framerate: bool = True,
    protect_controls: bool = True,
) -> None:
    """Load camera configuration.

    :param cam: The camera to load the configuration into.
    :param path: File path to the saved configuration
    :param controls: Names of the controls to load, if ``None`` then all are
      saved.
    :param format: Load the format.
    :param crop: Load the crop(s).
    :param framerate: Load the framerate.
    :param protect_controls: Protect sensor specific controls
      (ADC Gain, Black Level and VRAMP).

    :raises json.decoder.JSONDecodeError: If the json config fails to decode.
    """
    with open(path) as fd:
        data = fd.read()

    conf = json.loads(data)

    set(
        cam,
        conf,
        controls=controls,
        format=format,
        crop=crop,
        framerate=framerate,
        protect_controls=protect_controls,
    )


def reset_controls(
    cam: qamlib.Camera,
    controls: list[str] | None = None,
    protect_controls: bool = True,
) -> None:
    """Reset controls to their default value.

    :param cam: The camera to reset controls for.

    :param controls: The controls names to reset.

    :param protect_controls: Protect sensor specific controls
      (ADC Gain, Black Level and VRAMP).
    """
    ctrls: dict[str, qamlib.Control] = cam.list_controls()

    if controls is None:
        ctrl_names = list(ctrls.keys())
    else:
        ctrl_names = controls

    for c in ctrl_names:
        ctrl = ctrls[c]

        # This does not work for Trigger Sequence yet, so we skip it
        if c == "trigger sequence":
            continue

        if protect_controls and c in __PROTECED_CONTROLS:
            continue

        if ctrl.flags.write_only or ctrl.flags.read_only:
            continue

        val = cam.get_ext_control(c, default_value=True)

        cam.set_ext_control(c, val)
