"""High-level camera control API.

This API provides constructs for building high-level camera controls. The control
interface handles format conversions and validation of values automatically. The
API allows for hierarchical composition of controls for building organised control
structures.

"""

from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Callable
from typing import Generic
from typing import Optional
from typing import TypeVar

import numpy as np
import qamlib
from numpy import typing as npt

from quarklib.controls.formatters import AbstractFormatter
from quarklib.controls.formatters import ValidationStrategy
from quarklib.util import Comparable


T = TypeVar("T")
C = TypeVar("C", bound=Comparable)
TA = TypeVar("TA", bound=np.number)
U = TypeVar("U")


class IControl(Generic[T]):
    """Interface for camera controls.

    Provides methods for reading/writing/resetting the value as well as a few other
    convenience methods.

    :param cam: Camera used for connection.
    """

    name: str
    shape: tuple | dict
    cam: qamlib.Camera

    def __init__(self, cam: qamlib.Camera):
        self.cam = cam

    @abstractmethod
    def read(self) -> T:
        """Read the control value.

        :returns: The current value.
        """
        ...

    def write(
        self,
        value: T,
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ):  # noqa:
        """Write a new value to the control.

        :param value: Value to write.
        :param validation_strategy: Validate the value before writing. The
            default is to raise a `ValueError` on invalid values.
        """
        ...

    def reset(self):
        """Reset the control to its default value."""
        ...

    @property
    def formatter(self) -> Any:
        """Return the formatter of the control."""
        ...

    def get(self, item: str) -> Optional["IControl"]:
        """Get child if it exists."""
        return None


class DriverControl(ABC, IControl[T], Generic[T, C]):
    """Represents an actual control in the V4L2 driver.

    This kind of control automatically converts the raw values in the driver to and from
    a user specified format (e.g. float, enum, bool). Values are validated against the
    specified control format.

    :param cam: Camera object used for reading/writing the control.
    """

    _formatter: AbstractFormatter[C]
    default_value: T | None = None
    control_name: str
    id: int
    requires_event_handler: bool = False
    _qamlib_control: dict

    def __init__(
        self,
        cam: qamlib.Camera,
    ):
        super().__init__(cam)

    def __init_subclass__(cls, **kwargs):
        """Set up attributes for subclasses."""
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "_qamlib_control"):
            if cls.default_value is None:
                cls.default_value = cls._formatter.from_raw(
                    cls._qamlib_control["default_value"]
                )

    def __repr__(self):
        """Show the type of control and its formatter type."""
        return f"{self.name}: {self._formatter}"

    def control_callback(self, event: qamlib.ControlEvent):
        """Update the control state given an event."""
        pass

    @property
    def formatter(self) -> AbstractFormatter[C]:
        """Return the formatter associated with this control."""
        return self._formatter


class ScalarControl(DriverControl[C, C]):
    """A control that represents a single value."""

    shape = (1,)

    def read(self) -> C:
        """Read a scalar value from the control."""
        return self._formatter.from_raw(self.cam.get_control(self.control_name))

    def write(
        self,
        value: C,
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ):
        """Write a scalar value to the control."""
        raw = self._formatter.to_raw(value)
        self.cam.set_control(self.control_name, raw)

    def reset(self):
        """Reset the control to its default value."""
        self.cam.set_control(self.control_name, self._qamlib_control["default_value"])

    def __repr__(self):
        """Show the type of control, its formatter type, and its current value."""
        return f"{self.name} (scalar, {self._formatter}): {self.read()}"

    def control_callback(self, event: qamlib.ControlEvent):
        """Update the control's formatter."""
        self._formatter.update_info(event)


class ArrayControl(DriverControl[npt.NDArray[TA], TA]):
    """Control that represents an array of values.

    Array controls support the Numpy slicing syntax. Example:

    >>> cam = qamlib.Camera()
    >>> ctl = ArrayControl(
    ...     "lut_green",
    ...     cam,
    ...     cam.list_controls(),
    ...     FloatFormatter(3, 13),
    ... )
    >>> ctl[:25] = 2
    >>> print(ctl[:])
    >>> print(ctl.read())  # equivalent
    """

    _input_shape: tuple[int, ...]
    shape: tuple[int, ...]
    _from_raw: Callable[[npt.NDArray[np.integer]], npt.NDArray[TA]]
    _to_raw: Callable[[npt.NDArray[TA], ValidationStrategy], npt.NDArray[np.integer]]

    def __init_subclass__(cls, **kwargs):
        """Initialize subclasses.

        Sets up a number of attributes when creating new types.
        """
        if hasattr(cls, "_qamlib_control"):
            if not cls._qamlib_control.get("dimensions"):
                raise ValueError(
                    f"Incompatible type for {cls.control_name!r}. Expected "
                    f"an array but got a scalar value."
                )

            cls._from_raw = np.vectorize(cls._formatter.from_raw)
            cls._to_raw = np.vectorize(
                cls._formatter.to_raw,
                otypes=[np.uint32],
                excluded="validation_strategy",
            )

            dim: tuple[int, ...] = cls._qamlib_control["dimensions"]
            cls._input_shape = dim
            if len(dim) == 1:
                cls.shape = (dim[0], 1)
            else:
                cls.shape = dim

            if cls.default_value is None:
                cls.default_value = np.full(
                    dim,
                    cls._formatter.from_raw(cls._qamlib_control["default_value"]),
                )

    def read(self) -> npt.NDArray[TA]:
        """Read the control and ensure the output is correctly shaped."""
        raw_val = self.cam.get_ext_control(self.control_name).value.reshape(self.shape)

        return self._from_raw(raw_val)

    def write(
        self,
        value: npt.NDArray[TA],
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ):
        """Write a new value to the control.

        :param value: Value to write.
        :param validation_strategy: Validate the value before writing. The
            default is to raise a `ValueError` on invalid values.
        """
        raw = self._to_raw(value, validation_strategy).reshape(self._input_shape)
        self.cam.set_ext_control(self.control_name, qamlib.ArrayControlValue(raw))

    def reset(self):
        """Reset the control to its default value."""
        self.write(self.default_value)

    def __repr__(self):
        """Return information on the shape and formatter used for the array."""
        s = "x".join(str(o) for o in self.shape)
        return f"{self.name} (array {s}, {self._formatter}):\n{self.read()} "

    def __getitem__(self, item: Any) -> npt.NDArray[TA]:
        """Read a slice from the control.

        This method reads the entire control on access and is therefore not more
        efficient than a full read.

        :param item: The slice definition. Uses the NumPy syntax.
        :returns: The value of the slice.
        """
        return self.read()[item]

    def __setitem__(self, item: Any, value: npt.NDArray[TA]):
        """Write/modify a slice from the control.

        This method reads and writes the entire control on access and is therefore not
        more efficient than a full read and write.

        :param item: The slice definition. Uses the NumPy syntax.
        :param value: The value to write. The shape must be compatible under the
            NumPy broadcasting rules.
        """
        val = self.read()
        val[item] = value
        self.write(val)


class GroupControl(IControl[T]):
    """An abstract control for grouping other controls.

    This control does not have a driver equivalent. Instead, it is used to organize
    actual controls and to provide a simplified interface to multiple controls. It
    therefore contains several child controls.

    Groups allow reading/writing to either the entire group or to individual children.
    Each subtype defines its own method for accessing the children as a group.

    Writing to individual children is supported through attributes. The
    :class:`quarklib.camera.builders.GroupBuilder` helper class adds the children as
    attributes when creating object instances. This is achieved through subclassing
    such that each :class:`GroupControl` instance has its own type.

    :param cam: Camera object used for reading/writing the control.
    """

    _child_defs: list[type[IControl]]

    children: list[IControl]
    child_dict: dict[str, IControl]

    def __init__(self, cam: qamlib.Camera):
        super().__init__(cam)
        self.children = [c(cam) for c in self._child_defs]
        for c in self.children:
            setattr(self, f"_{c.name}", c)
        self.child_dict = {c.name: c for c in self.children}

    def __init_subclass__(cls, **kwargs):
        """Set the dimensions attribute."""
        super().__init_subclass__(**kwargs)

    @classmethod
    def _setup(cls):
        cls.shape = {c.name: c.shape for c in cls._child_defs}

    def reset(self):
        """Reset each of the child controls of the group."""
        for c in self.children:
            c.reset()

    def __repr__(self):
        """Return an info string of this group and its children."""
        children = [c.__repr__() for c in self.children]
        children = [
            "\n".join(["  " + line for line in s.splitlines()]) for s in children
        ]
        children = "\n".join(children)
        return f"{self.name}:\n{children}"

    @property
    def formatter(self) -> dict[str, AbstractFormatter[C] | dict]:
        """Return a dictionary of the formatters of each of this control's children."""
        return {k: v.formatter() for k, v in self.child_dict.items()}

    def get(self, item: str) -> IControl | None:
        """Get child if it exists."""
        return self.child_dict.get(item)


class DictGroupControl(GroupControl[dict[str, T]]):
    """Read/write children as dictionaries.

    When reading the group value, a dictionary containing all the children and
    their values is returned. When writing, the keys in the input dictionary
    determine which values will be written.
    """

    def read(self) -> dict[str, T]:
        """Read all child controls of the group.

        The values are returned as a dictionary with child names as keys.
        """
        return {c.name: c.read() for c in self.children}

    def write(
        self,
        value: dict[str, T],
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ):
        """Write multiple children simultaneously using a dictionary.

        :param value: A dictionary of child names mapped to values to write. The
            dictionary may contain all or some of the children.
        :param validation_strategy: Validate the value before writing. The
            default is to raise a `ValueError` on invalid values.
        """
        for name, val in value.items():
            self.child_dict[name].write(val, validation_strategy)


class GroupArrayControl(GroupControl[npt.NDArray[TA]]):
    """Allow multiple controls to be used as a single array.

    The control works by stacking read values into a single array and automatically
    splitting a written array into appropriate slices that are then written to each
    child control. For scalar children, the formed array is 1d. For 1d and 2d array
    controls, the formed array is a 2d array stacked horizontally.

    For example, an array group with three scalar children results in a 1x3 output
    shape. A group with two array children, with shapes 4x4 and 4x1 respectively,
    results in a 4x5 output shape.

    The array group can be sliced like a regular numpy array. Both reading and writing
    using slices is allowed. Slicing across control boundaries is also supported.
    For example, the color conversion matrix consists of a 4x4 matrix and a 4x1 matrix
    that can be grouped into a single 4x5 matrix using an array group. Slicing can be
    performed on this group without considering the control boundaries. The slightly
    different format conversions are also handled implicitly.

    >>> from qamlib import Camera
    >>> from quarklib.controls.builders import GroupControlDefinition, \
ControlDefinition
    >>> color_def = GroupControlDefinition.array(
    ...        "ccm",
    ...        children=[
    ...            ControlDefinition.float_array("matrix", "color_conversion_matrix",
    ...                 Q_3_12),
    ...            ControlDefinition.float_array("offset", "color_conversion_offset",
    ...                 Q_0_14),
    ...        ],
    ...    ),
    >>> cam = Camera()
    >>> with cam:
    >>>     color = color_def.build_controls(cam)
    >>>     color[2:3, 3:5] = np.ones((1, 2))
    """

    _dim_stops: list[int]
    _l_dim: int

    @classmethod
    def _setup(cls):
        super()._setup()
        # Determines the index positions of the boundaries of each child.
        # These are used to split written arrays into the necessary parts.
        shapes = [c.shape for c in cls._child_defs]
        if all(len(s) == 1 for s in shapes):
            cls._l_dim = 0
        else:
            cls._l_dim = 1

        sh = list(shapes[0])
        sh[cls._l_dim] = sum(s[cls._l_dim] for s in shapes)
        cls.shape = tuple(sh)

        cls._dim_stops = []
        current_x = 0
        for s in shapes:
            # print(s)
            cls._dim_stops.append(current_x)
            current_x += s[cls._l_dim]
        cls._dim_stops.append(current_x)

    def read(self) -> npt.NDArray[TA]:
        """Read children and stack the results horizontally into a single array.

        :returns: The stacked array of values.
        """
        val_list = [c.read() for c in self.children]
        return np.hstack(val_list)

    def write(
        self,
        value: npt.NDArray[TA],
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ):
        """Write a value to all children using a single array.

        The array value is split according to the shape of each child and each slice is
        written to the appropriate control. It essentially mirrors the :meth:`read`
        method.

        :param value: Array to be written.
        :param validation_strategy: Validate the value before writing. The
            default is to raise a `ValueError` on invalid values.
        """
        value = np.array(value)
        val_list = [
            value.take(
                indices=range(self._dim_stops[i], self._dim_stops[i + 1]),
                axis=self._l_dim,
            )
            for i in range(len(self.children))
        ]
        for val, c in zip(val_list, self.children, strict=True):
            c.write(val, validation_strategy)

    def __setitem__(self, item: list[slice], value: npt.NDArray[TA]):
        """Write/modify a slice from the control.

        This method reads and writes the entire control on access and is therefore not
        more efficient than a full read and write.

        :param item: The slice definition. Uses the NumPy syntax.
        :param value: The value to write. The shape must be compatible under the
            NumPy broadcasting rules.
        """
        val = self.read()
        val[item] = value
        self.write(val)

    def __getitem__(self, item: Any) -> npt.NDArray[TA]:
        """Read a slice from the control.

        This method reads the entire control on access and is therefore not more
        efficient than a full read.

        :param item: The slice definition. Uses the NumPy syntax.
        :returns: The value of the slice.
        """
        return self.read()[item]


class GroupSplitControl(GroupControl[list[T]]):
    """Write the same value to multiple controls.

    This group makes it easy to demonstrate the intent to replicate the same value
    across multiple controls. A single call to `write` will write the same value to
    all children.

    The control requires that all children have the same shape and value types.
    """

    def write(
        self,
        value: T | list[T],
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ):
        """Write the same value to all children."""
        if isinstance(value, list):
            for c, v in zip(self.children, value, strict=True):
                c.write(v, validation_strategy)
        else:
            for c in self.children:
                c.write(value, validation_strategy)

    def read(self) -> list[T]:
        """Read all child controls.

        This method still reads each child separately, since their values are not
        guaranteed to be the same.
        """
        return [c.read() for c in self.children]
