"""Defines helpers for building camera controls."""

import logging
from typing import Any
from typing import Protocol
from typing import cast

import numpy as np
import numpy.typing as npt
import qamlib

from quarklib.controls import ArrayControl
from quarklib.controls import DictGroupControl
from quarklib.controls import DriverControl
from quarklib.controls import GroupArrayControl
from quarklib.controls import GroupControl
from quarklib.controls import GroupSplitControl
from quarklib.controls import IControl
from quarklib.controls import ScalarControl
from quarklib.controls.formatters import AbstractFormatter
from quarklib.controls.formatters import BoolFormatter
from quarklib.controls.formatters import FloatFormatter
from quarklib.controls.formatters import IntFormatter
from quarklib.controls.formatters import MappingFormatter
from quarklib.controls.formatters import MenuFormatter
from quarklib.util import cam_info_to_dict
from quarklib.util import to_snake_case


log = logging.getLogger("quarklib.controls")


class IControlDefinition(Protocol):  # noqa: F811
    """Helper classes for building :class:`IControl` objects."""

    name: str
    doc: str
    optional: bool
    kind: type[IControl]

    def build_types(self, info: dict[str, dict], parent: type) -> type[IControl]:
        """Build control from camera and extracted info dict.

        :param info: Camera info (from `qamlib.Camera.listcontrols`)
        :param parent: Parent object
        :returns: The built control object.
        """
        pass


class ControlDefinition(IControlDefinition):
    """Builds :class:`DriverControl` type controls.

    The class contains a number of helper methods for simplifying the building of
    common control types.

    The builder creates a new type with the name of the control when building.

    :param name: Name of the control.
    :param formatter: Control formatter type.
    :param control: Control type.
    :param control_name: Actual control name.
    :param formatter_args: Keyword arguments for the formatter constructor.
    :param default_value: Default control value.
    :param optional: Whether the control is expected to always be present.
    :param requires_event_handler: `True` if the control should always
        receive update events.
    :param doc: Documentation string inserted into the created type.
    """

    def __init__(
        self,
        name: str,
        formatter: type[AbstractFormatter],
        control: type[DriverControl],
        control_name: str | None = None,
        formatter_args: dict | None = None,
        default_value: Any = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        self.name = name
        self.control_name = (
            to_snake_case(name) if control_name is None else control_name
        )
        self.formatter_args = {} if formatter_args is None else formatter_args
        self.control = control
        self.formatter = formatter
        self.doc = doc
        self.__doc__ = doc
        self.kind = DriverControl
        self.optional = optional
        self.requires_event_handler = requires_event_handler
        self.default_value = default_value

    @classmethod
    def float_array(
        cls,
        name: str,
        control_name: str | None = None,
        formatter_args: dict | None = None,
        default_value: npt.NDArray[np.floating] | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create a floating point array control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param formatter_args: Keyword arguments for the formatter constructor.
        :param default_value: Default control value.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: `True` if the control should always
            receive update events.
        :param doc: Documentation string inserted into the created type.

        :returns: The constructed control.
        """
        return cls(
            name,
            formatter=FloatFormatter,
            control=ArrayControl,
            formatter_args=formatter_args,
            control_name=control_name,
            default_value=default_value,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    @classmethod
    def float_scalar(
        cls,
        name: str,
        control_name: str | None = None,
        formatter_args: dict | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create a floating point scalar control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param formatter_args: Keyword arguments for the formatter constructor.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: Whether the control should always receive
            update events.
        :param doc: Documentation string inserted into the created type.
        :returns: The constructed controller.
        """
        return cls(
            name,
            formatter=FloatFormatter,
            control=ScalarControl,
            formatter_args=formatter_args,
            control_name=control_name,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    @classmethod
    def mapped_scalar(
        cls,
        name: str,
        control_name: str | None = None,
        formatter_args: dict | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create a floating point scalar control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param formatter_args: Keyword arguments for the formatter constructor.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: Whether the control should always receive
            update events.
        :param doc: Documentation string inserted into the created type.
        :returns: The constructed controller.
        """
        return cls(
            name,
            formatter=MappingFormatter,
            control=ScalarControl,
            formatter_args=formatter_args,
            control_name=control_name,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    @classmethod
    def boolean(
        cls,
        name: str,
        control_name: str | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create a boolean scalar control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: `True` if the control should always
            receive update events.
        :param doc: Documentation string inserted into the created type.
        :returns: The constructed controller.
        """
        return cls(
            name,
            formatter=BoolFormatter,
            control=ScalarControl,
            control_name=control_name,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    @classmethod
    def int_array(
        cls,
        name: str,
        control_name: str | None = None,
        formatter_args: dict | None = None,
        default_value: npt.NDArray[np.integer] | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create an integer array control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param formatter_args: Keyword arguments for the formatter constructor.
        :param default_value: Default control value.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: Whether the control should always receive
            update events.
        :param doc: Documentation string inserted into the created type.
        :returns: The constructed controller.
        """
        return cls(
            name,
            formatter=IntFormatter,
            control=ArrayControl,
            formatter_args=formatter_args,
            control_name=control_name,
            default_value=default_value,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    @classmethod
    def int_scalar(
        cls,
        name: str,
        control_name: str | None = None,
        formatter_args: dict | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create an integer scalar control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param formatter_args: Keyword arguments for the formatter constructor.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: Whether the control should always receive
            update events.
        :param doc: Documentation string inserted into the created type.
        :returns: The constructed control.
        """
        return cls(
            name,
            formatter=IntFormatter,
            control=ScalarControl,
            formatter_args=formatter_args,
            control_name=control_name,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    @classmethod
    def menu(
        cls,
        name: str,
        control_name: str | None = None,
        optional: bool = False,
        requires_event_handler: bool = False,
        doc: str = "",
    ):
        """Create a menu scalar control.

        :param name: Control name.
        :param control_name: Actual control name.
        :param optional: Whether the control is expected to always be present.
        :param requires_event_handler: Whether the control should always receive
            update events.
        :param doc: Documentation string inserted into the created type.

        :returns: The constructed control.
        """
        return cls(
            name,
            formatter=MenuFormatter,
            control=ScalarControl,
            control_name=control_name,
            optional=optional,
            requires_event_handler=requires_event_handler,
            doc=doc,
        )

    def build_types(self, info: dict[str, dict], parent: type) -> type[IControl]:
        """Build the control.

        Creates a new class based on the specified base class and with the specified
        parameters. Some attributes are inferred from the `info` dictionary.
        """
        try:
            qamlib_control = info[self.control_name]
        except KeyError:
            raise OSError(
                f"The control with name {self.control_name!r} is either not supported "
                "by the camera or has been misspelled."
            ) from None

        # Create a new type for this specific control.
        new_type = type(
            self.name,
            (self.control,),
            {
                "name": to_snake_case(self.name),
                "_formatter": self.formatter(**self.formatter_args, **qamlib_control),
                "control_name": self.control_name,
                "default_value": self.default_value,
                "id": qamlib_control["id"],
                "requires_event_handler": self.requires_event_handler,
                "_qamlib_control": qamlib_control,
                "__module__": parent.__name__,
                "__doc__": self.doc,
                "__annotations__": {
                    "formatter": self.formatter.__name__,
                },
            },
        )
        setattr(parent, self.name, new_type)

        return new_type


class GroupControlDefinition(IControlDefinition):
    """Builds :class:`GroupControl` type controls.

    This builder builds child controls recursively, thus allowing for a single build
    call to construct a large hierarchy of camera controls.

    When building groups, the specific :class:`GroupControl` subclass is subclassed
    again to create a new type that includes the specified children as attributes.

    :param name: Name of the control.
    :param kind: Group kind to construct.
    :param children: Child controls. Can be either groups or driver controls.
    :param optional: Whether the control is expected to always be present.
    :param doc: Documentation string inserted into the created type.
    """

    def __init__(
        self,
        name: str,
        kind: type[GroupControl],
        children: list[IControlDefinition],
        optional: bool = False,
        doc: str = "",
    ):
        self.name = name
        self.children = children
        self.kind = kind
        self.optional = optional
        self.doc = doc

        child_docs = [f"- {c.name}: {c.doc}" for c in children]
        newline = "\n"
        self.__doc__ = f"""Control definition for {self.name}

Description: {self.doc}

Children:

{newline.join(child_docs)}
        """

    @classmethod
    def basic(
        cls,
        name: str,
        children: list[IControlDefinition],
        optional: bool = False,
        doc: str = "",
    ):
        """Build a regular group control.

        :param name: Control name.
        :param children: Control children.
        :param optional: Whether the control is expected to always be present.
        :param doc: Optional type documentation.
        :returns: The constructed control.
        """
        return cls(name, DictGroupControl, children, optional, doc)

    @classmethod
    def array(
        cls,
        name: str,
        children: list[IControlDefinition],
        optional: bool = False,
        doc: str = "",
    ):
        """Build an array group.

        :param name: Control name.
        :param children: Control children.
        :param optional: Whether the control is expected to always be present.
        :param doc: Optional type documentation.
        :returns: The constructed control.
        """
        return cls(name, GroupArrayControl, children, optional, doc)

    @classmethod
    def split(
        cls,
        name: str,
        children: list[IControlDefinition],
        optional: bool = False,
        doc: str = "",
    ):
        """Build a split group.

        :param name: Control name..
        :param children: Control children.
        :param optional: Whether the control is expected to always be present.
        :param doc: Optional type documentation.
        :returns: The constructed control.
        """
        return cls(name, GroupSplitControl, children, optional, doc)

    def build_types(self, info: dict[str, dict], parent: type) -> type[GroupControl]:
        """Build the control and all its children.

        Creates a new class based on the specified base class and with the specified
        parameters. Some attributes are inferred from the `info` dictionary.

        All child control types are associated with the parent control's class.
        """
        new_type = cast(
            type[DictGroupControl],
            type(
                self.name,
                (self.kind,),
                {
                    "name": to_snake_case(self.name),
                    "__doc__": self.doc,
                    "__module__": parent.__name__,
                },
            ),
        )
        child_types: list[type[IControl]] = []
        for c in self.children:
            try:
                c_type = c.build_types(info, new_type)
                child_types.append(c_type)

                def prop_gen(obj: Any, child: IControlDefinition = c) -> IControl:
                    return getattr(obj, f"_{to_snake_case(child.name)}")

                prop_gen.__annotations__["return"] = c.kind

                setattr(
                    new_type,
                    to_snake_case(c.name),
                    property(prop_gen, doc=c.doc),
                )
            except OSError as e:
                if c.optional:
                    log.info(e)
                else:
                    raise e

        children = {c.name: c.__name__ for c in child_types}
        new_type._child_defs = child_types
        new_type.__annotations__ = children
        # for c_type in child_types:
        #     setattr(new_type, cast(type, c_type).__name__, c_type)

        # setattr(parent, self.name, new_type)

        new_type._setup()
        # new_type.dimensions = [c.dimensions for c in new_type._child_defs]

        return new_type

    def build_controls(
        self, cam: qamlib.Camera, parent: type | None = None
    ) -> GroupControl:
        """Build control from camera and extracted info dict.

        :param cam:
        :param parent:
        :returns: The built control object.
        """
        if parent is None:
            parent = type(self)
        tp = self.build_types(cam_info_to_dict(cam.list_controls()), parent)
        return tp(cam)
