"""Defines formatter types to be used in camera controls."""

import difflib
from abc import ABC
from abc import abstractmethod
from enum import Enum
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar
from typing import cast

import numpy as np
import qamlib

from quarklib.util import Comparable
from quarklib.util import from_fixedpoint
from quarklib.util import to_fixedpoint


T = TypeVar("T", bound=Comparable)
U = TypeVar("U")


def _clamp_value(value: T, min_value: T, max_value: T) -> T:
    return max(min(value, max_value), min_value)


class ValidationStrategy(Enum):
    """Validation strategies.

    Determines how validation is handled by a formatter.
    """

    Error = "Error"
    """Raise a ValueError if the provided value is outside the valid set."""
    ConvertClamp = "ConvertClamp"
    """Attempt to convert the provided value to the correct type and clamp it
    to the closest value in the valid set.
    """


class Interval(Generic[T]):
    """Represents a half-open interval.

    :param lower: Lower bound (inclusive).
    :param upper: Upper bound (inclusive).
    :param step: Interval step.
    """

    def __init__(self, lower: T, upper: T, step: T):
        self.lower = lower
        self.upper = upper
        self.step = step

    def __contains__(self, item: T) -> bool:
        """Test if item is contained in the interval.

        :param item: Value to compare.
        :returns: True if the item is contained in the interval, False otherwise.
        """
        return self.lower <= item < self.upper

    def __repr__(self) -> str:
        """Show the valid interval."""
        return f"[{self.lower}, {self.upper}, {self.step})"


class AbstractFormatter(ABC, Generic[T]):
    """The base class for all formatters.

    A formatter is a type that can convert to and from a raw integer value.
    """

    _type: type

    def __init__(self, **kwargs):
        pass

    def to_raw(
        self,
        value: T,
        validation_strategy: ValidationStrategy = ValidationStrategy.Error,
    ) -> int:
        """Convert a value to integer with validation.

        The validation procedure is implemented by each subclass and should prevent
        invalid values from being successfully converted. In other words, returned
        values should always represent valid values.

        :param value: The input value.
        :param validation_strategy: How to handle validation. The default is to raise
            a `ValueError` on invalid values.

        :raises ValueError: If `validate` is True and there is a validation error.

        :returns: The raw integer value.
        """
        match validation_strategy:
            case ValidationStrategy.Error:
                self.validate_value(value)
            case ValidationStrategy.ConvertClamp:
                value = self.validate_value(value)
        return self._to_raw(value)

    @abstractmethod
    def _to_raw(self, value: T) -> int:
        pass

    @abstractmethod
    def from_raw(self, raw: int) -> T:
        """Convert a raw integer to a user value.

        This method does not check the validity of the input since all written values
        have already been verified.
        """
        pass

    @property
    @abstractmethod
    def raw_valid_set(self) -> Interval | set:
        """Returns the range of valid raw values.

        The returned value can either be a range describing an interval of valid values
        or a set of valid values.
        """
        pass

    @property
    @abstractmethod
    def valid_set(self) -> Interval | set:
        """Returns the range of valid user values.

        The returned value can either be a range describing an interval of valid values
        or a set of valid values.
        """
        pass

    def update_info(self, event: qamlib.ControlEvent):
        """Update formatter info based on the received event."""
        pass

    def validate_value(self, value: T):
        """Validate a user value.

        The default implementation checks if the value is in the set of valid values.

        :param value: User value to validate.
        :raises ValueError: Raised if the value is invalid.
        """
        if value not in self.valid_set:
            raise ValueError(
                f"Value {value} of type {type(value).__name__} is outside the allowed "
                f"range of {self.valid_set}"
            )

    @abstractmethod
    def clamp_value(self, value: Any) -> T:
        """Attempt to convert the given value to a valid element of the valid set."""
        pass

    def __repr__(self):
        """Return formatter type as string."""
        return str(type(self).__name__)

    def __str__(self):
        """Return formatter type as string."""
        return self.__repr__()


class BoolFormatter(AbstractFormatter[bool]):
    """Formatter of boolean values."""

    _type = bool

    def __init__(self, **kwargs):
        super().__init__(raw_type=int)

    def _to_raw(self, value: bool) -> int:
        return int(value)

    def from_raw(self, raw: int) -> bool:
        """Convert from integer to bool."""
        return bool(raw)

    @property
    def raw_valid_set(self) -> Interval | set:
        """The valid raw values (0 or 1)."""
        return {0, 1}

    @property
    def valid_set(self) -> Interval | set:
        """The valid user values (True or False)."""
        return {False, True}

    def validate_value(self, value: bool):
        """Values are valid if they have type bool."""
        if not isinstance(value, bool):
            raise ValueError(f"Invalid value {value} for boolean property.")
        else:
            super().validate_value(value)

    def clamp_value(self, value: Any) -> bool:
        """Convert the value to bool."""
        return bool(value)


class IntFormatter(AbstractFormatter[int]):
    """Formatter of integers.

    This formatter doesn't modify the value itself since it can be written directly to
    the V4L2 control. Instead, it only performs validation.
    """

    _type = int

    def __init__(
        self,
        min: int,
        max: int,
        dtype: Callable[[Any], np.integer] = np.int32,
        **kwargs,
    ):
        self.dtype = dtype
        self._min = min
        self._max = max
        super().__init__(**kwargs)

    def _to_raw(self, value: int) -> int:
        return int(self.dtype(value))

    def from_raw(self, raw: int) -> int:
        """Return the raw value which is the same as the user value."""
        return raw

    @property
    def raw_valid_set(self) -> Interval:
        """The valid set is an interval between min and max values."""
        return Interval(self._min, self._max, 1)

    @property
    def valid_set(self) -> Interval:
        """The valid set is an interval between min and max values."""
        return self.raw_valid_set

    def clamp_value(self, value: Any) -> int:
        """Attempt to convert the value to an integer in the valid set."""
        return _clamp_value(int(value), self.valid_set.lower, self.valid_set.upper)

    def update_info(self, event: qamlib.ControlEvent):
        """Set the min/max values for the control based on the values in `event`."""
        self._min = event.min
        self._max = event.max

    def __repr__(self):
        """Show valid interval."""
        return f"Int [{self.valid_set.lower}, {self.valid_set.upper}]"


class MappingFormatter(AbstractFormatter[float]):
    """Custom mapping formatter."""

    _type = float

    def __init__(
        self,
        min,
        max,
        forward_map: Callable[[int], float],
        backward_map: Callable[[float], int],
        **kwargs,
    ):
        self._min = min
        self._max = max
        self.forward_map = forward_map
        self.backward_map = backward_map
        super().__init__(**kwargs)

    def _to_raw(self, value: float) -> int:
        return self.backward_map(value)

    def from_raw(self, raw: int) -> float:
        """Return the raw value which is the same as the user value."""
        return self.forward_map(raw)

    @property
    def raw_valid_set(self) -> Interval:
        """The valid set is an interval between min and max values."""
        return Interval(self._min, self._max, 1)

    @property
    def valid_set(self) -> Interval:
        """The valid set is an interval between min and max values."""
        # The step parameter is only correct if the mapper is linear.
        return Interval(
            self.forward_map(self._min),
            self.forward_map(self._max),
            self.forward_map(1) - self.forward_map(0),
        )

    def clamp_value(self, value: Any) -> int:
        """Attempt to convert the value to an integer in the valid set."""
        return _clamp_value(int(value), self.valid_set.lower, self.valid_set.upper)

    def update_info(self, event: qamlib.ControlEvent):
        """Set the min/max values for the control based on the values in `event`."""
        self._min = event.min
        self._max = event.max

    def __repr__(self):
        """Show valid interval."""
        return (
            f"Mapper [{self.raw_valid_set.lower} -> {self.valid_set.lower},"
            " {self.raw_valid_set.upper} -> {self.valid_set.upper}]"
        )


class FloatFormatter(AbstractFormatter[float]):
    """Formatter for floating point values.

    This formatter converts to/from fixed point numbers that are used for V4L2 controls.
    The formatter specifies the fixed-point format and always rounds to the nearest
    representable value when converting from floating point.

    The validation uses only an interval between min and max values and thus does not
    check if incoming values can be represented exactly as the fixed point format.
    """

    _type = float

    def __init__(self, int_bits=0, frac_bits=0, signed=False, **kwargs):
        super().__init__(**kwargs)
        self.int_bits = int_bits
        self.frac_bits = frac_bits
        self.signed = signed
        self.np_type = np.int32 if signed else np.uint32

    def _to_raw(self, value: float) -> int:
        return self.np_type(
            to_fixedpoint(value, self.int_bits, self.frac_bits, self.signed)
        )

    def from_raw(self, raw: int) -> float:
        """Convert from fixed point value to floating point value."""
        return from_fixedpoint(raw, self.int_bits, self.frac_bits, self.signed)

    @property
    def raw_valid_set(self) -> Interval:
        """The (continuous) interval of valid raw integer values."""
        total_bits = self.int_bits + self.frac_bits
        return Interval(0, 2**total_bits, 1)

    @property
    def valid_set(self) -> Interval:
        """The interval of valid floating point numbers."""
        if self.signed:
            base = 2 ** (self.int_bits - 1)
            return Interval(-base, base - 2 ** (-self.frac_bits), self.precision)
        else:
            return Interval(
                0.0, 2**self.int_bits - 2 ** (-self.frac_bits), self.precision
            )

    def clamp_value(self, value: Any) -> float:
        """Attempt to convert the value to a float in the valid set."""
        return _clamp_value(float(value), self.valid_set.lower, self.valid_set.upper)

    @property
    def precision(self) -> float:
        """The distance between two consecutive numbers of the fixed point format."""
        return 2 ** (-self.frac_bits)

    def __repr__(self):
        """Show the type information and limits."""
        return (
            f"Q{self.int_bits}.{self.frac_bits} [{self.valid_set.lower:.1f}, "
            f"{self.valid_set.upper:.1f}), precision: {self.precision:.2e}"
        )


class MenuFormatter(AbstractFormatter[str]):
    """Formatter for menu values (enumerable types).

    The formatter uses the string values as user values and their corresponding integer
    indices as raw values.
    """

    _type = str

    def __init__(self, items, **kwargs):
        super().__init__(**kwargs)
        self._menu_items = list(items.values())

    def _to_raw(self, value: str) -> int:
        try:
            return self._menu_items.index(value)
        except ValueError:
            raise ValueError(
                f"Invalid menu value {value!r}. Allowed values: {self._menu_items}"
            ) from ValueError

    def from_raw(self, raw: int) -> str:
        """Get the menu item for the input index."""
        return self._menu_items[raw]

    @property
    def raw_valid_set(self) -> Interval | set:
        """Return the set of valid indices."""
        return set(range(len(self._menu_items)))

    @property
    def valid_set(self) -> Interval | set:
        """Return the set of valid string values."""
        return set(self._menu_items)

    def validate_value(self, value: str):
        """Check if the provided string is contained in the valid set."""
        if value not in self.valid_set:
            assert isinstance(self.valid_set, set)
            raise ValueError(
                f"Menu value {value} is outside the allowed range of "
                f"({'|'.join(self.valid_set)})"
            )

    def clamp_value(self, value: Any) -> str:
        """Attempt to convert the value to a `str`.

        Find the closes valid menu string using `difflib.get_close_matches`.
        """
        return difflib.get_close_matches(
            str(value), cast(set, self.valid_set), n=1, cutoff=0
        )[0]

    def __repr__(self):
        """Show menu items."""
        assert isinstance(self.valid_set, set)
        inner = ", ".join(self.valid_set)  # type: ignore[arg-type]
        return f"menu -> {inner}"
