"""Contains the base control definitions used in this library."""

import math

import numpy as np

from quarklib.controls.builders import ControlDefinition
from quarklib.controls.builders import GroupControlDefinition


Q_0_12 = {"int_bits": 0, "frac_bits": 12, "signed": False}
Q_3_12 = {"int_bits": 3, "frac_bits": 13, "signed": True}
Q_0_14 = {"int_bits": 0, "frac_bits": 12, "signed": True}
Q_2_14 = {"int_bits": 2, "frac_bits": 14, "signed": False}
Q_2_14_signed = {"int_bits": 2, "frac_bits": 14, "signed": True}

LUT = GroupControlDefinition.basic(
    "Lut",
    optional=True,
    doc="Lookup Table Controls",
    children=[
        ControlDefinition.float_array(
            "Mono", "lut_mono", Q_0_12, doc="Lookup table for grayscale formats."
        ),
        GroupControlDefinition.array(
            "Rgb",
            children=[
                ControlDefinition.float_array(
                    "Red",
                    "lut_red",
                    Q_0_12,
                    doc="Lookup table for the red color channel.",
                ),
                ControlDefinition.float_array(
                    "Green",
                    "lut_green",
                    Q_0_12,
                    doc="Lookup table for the green color channel.",
                ),
                ControlDefinition.float_array(
                    "Blue",
                    "lut_blue",
                    Q_0_12,
                    doc="Lookup table for the blue color channel.",
                ),
            ],
            doc="Combined lookup tables for RGB.",
        ),
    ],
)
""""""

COLOR_CONVERSION = GroupControlDefinition.basic(
    "ColorConversion",
    optional=True,
    children=[
        ControlDefinition.boolean("Enable", "enable_color_conversion"),
        ControlDefinition.int_scalar("SaturationValue"),
        GroupControlDefinition.array(
            "Ccm",
            children=[
                ControlDefinition.float_array(
                    "Matrix",
                    "color_conversion_matrix",
                    Q_3_12,
                    default_value=np.eye(4),
                    doc="Linear color transformation using the RGBA format.",
                ),
                ControlDefinition.float_array(
                    "Offset",
                    "color_conversion_offset",
                    Q_0_14,
                    doc="Color translation.",
                ),
            ],
            doc="Affine color conversion transform.",
        ),
    ],
    doc="Color Conversion Controls",
)
""""""

DIGITAL_GAIN = GroupControlDefinition.array(
    "DigitalGain",
    children=[
        ControlDefinition.float_scalar(
            "Red", formatter_args=Q_2_14, control_name="red_balance"
        ),
        ControlDefinition.float_scalar(
            "Green", formatter_args=Q_2_14, control_name="green_balance"
        ),
        ControlDefinition.float_scalar(
            "Blue", formatter_args=Q_2_14, control_name="blue_balance"
        ),
    ],
    doc="Digital Gain Controls, usually used for white balance.",
)
""""""

FLIP = GroupControlDefinition.basic(
    "Flip",
    optional=True,
    children=[
        ControlDefinition.boolean(
            "Horizontal", "horizontal_flip", doc="Flip image horizontally."
        ),
        ControlDefinition.boolean(
            "Vertical", "vertical_flip", doc="Flip image vertically."
        ),
    ],
    doc="Flip Controls.",
)
""""""

BINNING = GroupControlDefinition.basic(
    "Binning",
    children=[
        ControlDefinition.int_scalar("Horizontal", "horizontal_binning"),
        ControlDefinition.int_scalar("Vertical", "vertical_binning"),
    ],
    doc="Camera binning controls.",
)
""""""

TRIGGER = GroupControlDefinition.basic(
    "Trigger",
    children=[
        ControlDefinition.menu("Mode", control_name="trigger_mode"),
        ControlDefinition.boolean(
            "InvertPolarity",
            "invert_trigger_polarity",
        ),
        ControlDefinition.int_scalar("ExternalDelay", "external_trigger_delay"),
    ],
    doc="Trigger related controls.",
)
""""""

FLASH = GroupControlDefinition.basic(
    "Flash",
    children=[
        ControlDefinition.boolean(
            "InvertPolarity",
            "invert_flash_polarity",
        ),
        ControlDefinition.boolean("Disable", "disable_flash"),
    ],
    doc="Flash related controls.",
)
""""""

GAIN_MAP = GroupControlDefinition.basic(
    "GainMap",
    optional=True,
    children=[
        ControlDefinition.boolean("Enable", "gain_map"),
        ControlDefinition.float_scalar("Scale", "xform_gain_scale", Q_2_14),
        ControlDefinition.int_scalar(
            "Offset", "extra_gain_for_gain_map", {"dtype": np.uint8}
        ),
        ControlDefinition.int_array(
            "Map", "gain_map_data", {"dtype": np.uint8}, doc="Map data array."
        ),
    ],
    doc="Xform gain map controls.",
)
""""""

PATTERN_GENERATOR = ControlDefinition.menu(
    "PatternGenerator", doc="Set generated pattern (for IMX sensors).", optional=True
)
""""""

CAPTURE_OPTIONS = GroupControlDefinition.basic(
    "Capture",
    children=[
        ControlDefinition.int_scalar(
            "ExposureTime",
            control_name="exposure_time_absolute",
            formatter_args={"dtype": np.uint16},
            requires_event_handler=True,
            optional=True,
        ),
        ControlDefinition.mapped_scalar(
            "AnalogGain",
            control_name="gain",
            formatter_args={
                "forward_map": lambda x: 10 ** (x / 1000 / 20),
                "backward_map": lambda x: int(1000 * 20 * math.log10(x)),
            },
            optional=True,
        ),
    ],
)

EXPOSURE_SEQUENCE = GroupControlDefinition.basic(
    "ExposureSequence",
    doc="Controls for exposure sequence settings for the GSENSE sensor line.",
    optional=True,
    children=[
        ControlDefinition.int_array(
            "LineExposureSequence",
            formatter_args={"dtype": np.uint32},
            doc="""Line exposure sequence.

Define exposure per line. Exposure is always defined on the full 2048 lines
irrespective of any active cropping area.""",
        ),
        ControlDefinition.int_array(
            "LineOrder",
            formatter_args={"dtype": np.uint32},
            doc="""Line ordering.
Determines which lines to select for the final frame. Its size depends on the
current cropping height.""",
            requires_event_handler=True,
        ),
    ],
)

QTEC_DEFAULT_CONTROLS = GroupControlDefinition.basic(
    "QtecDefaultControls",
    [
        DIGITAL_GAIN,
        GAIN_MAP,
        COLOR_CONVERSION,
        LUT,
        FLIP,
        BINNING,
        PATTERN_GENERATOR,
        CAPTURE_OPTIONS,
        EXPOSURE_SEQUENCE,
    ],
)
