"""Save functions."""

import sys

import numpy as np


def needs_byteswap(frame: np.ndarray):
    """Check if a byte-swap is needed for the PNM format."""
    # 8-bit
    if frame.itemsize == 1:
        return False
    # Big Endian
    if frame.dtype.byteorder == ">":
        return False
    # Big Endian native byte order
    if frame.dtype.byteorder == "=" and sys.byteorder == "big":
        return False
    return True


def save_pnm(name: str, frame: np.ndarray) -> None:
    """Save frame as PNM file.

    NOTE: For 16 bit frames PNM uses big endian, so we may have to do a
    byte-swap, this can decrease performance, but it is possible to improve the
    byte-swap speed by doing it in-place ``frame =
    frame.byteswap(True).newbyteorder()`` before calling :func:`save_pnm`

    :param name: The file name to save to.

    :param frame: The frame to save.
    """
    with open(name, "wb") as f:
        height, width, channels = frame.shape

        if channels == 1:
            type = "P5"
        else:
            type = "P6"

        if frame.nbytes == width * height * channels:
            if frame.dtype == "uint8":
                max_val = 255
            else:
                max_val = 65535
        else:
            max_val = 65535

        pnm_header = f"{type} {width} {height} {max_val}\n"
        f.write(bytearray(pnm_header, "ascii"))

        if needs_byteswap(frame):
            frame.byteswap().tofile(f)
        else:
            frame.tofile(f)
