"""Module containing classes for doing auto exposure."""

import logging
import time
from typing import cast

import cv2
import numpy
import numpy as np
import numpy.typing as npt
import qamlib


# Get quarklib logger
log = logging.getLogger("quarklib")


class AutoExposure:
    """Auto exposure base class.

    This is the base class for auto exposure classes. The classes only support
    certain known formats: ``BGR3``, ``BGR4``, ``HSV``, ``RGB3``, ``RGB4``,
    ``GREY``, ``GREY16``.

    :param cam: The ``qamlib.Camera`` object to do the auto exposure on.
    :param crop: The crop of the frame to focus the auto exposure algorithm
      on.
    :param num_average: The number of frames to average over before
      calculating a new exposure.
    """

    def __init__(
        self,
        cam: qamlib.Camera,
        *,
        crop: tuple[int, int, int, int] | None,
        num_average: int,
    ) -> None:
        self.cam = cam
        self.num_average = num_average

        self._crop: tuple[slice, slice] | None = None
        self._avg_frame: npt.NDArray[numpy.float64] | None = None
        self._n_frames = 0
        self.last_exposure_change = 0.0

        self.update_settings()

        if crop is not None:
            s1 = slice(crop[1], crop[1] + crop[3])  # Width crop
            s2 = slice(crop[0], crop[0] + crop[2])  # Height crop
            self._crop = (s1, s2)

    def update_settings(self) -> None:
        """Get updated camera settings used for auto exposure.

        Such as maximum exposure, and current format. This function needs to be
        called if framerate, or format has been changed after creating the object.
        """
        self._format = self.cam.get_format()

        self._exposure = self.cam.list_controls()["exposure_time_absolute"]

        self._warned_exposure = False

    def iterate(
        self, meta: qamlib.FrameMetadata, frame: npt.NDArray[numpy.uint8]
    ) -> float | None:
        """Do one iteration of auto exposure.

        This will only change exposure every ``num_average`` frames.

        NOTE: This function is only valid for sub-classes.
        """
        raise Exception(
            "Not valid to call iterate on a AutoExposure object, use one of the"
            " inherietors, eg. SimpleAutoExposure"
        )


class SimpleAutoExposure(AutoExposure):
    """Simple auto exposure.

    This auto exposure class will try to saturate the image up to
    ``target_value``. When increasing the exposure it uses the maximum pixel
    value in the frame to calculate how much to increase exposure.

    :param cam: :class:`qamlib.Camera` object to do auto exposure on.
    :param crop: The part of the frame to look at for auto exposure (left, top,
      width, height)
    :param num_average: Number of frames to average before calculating new
      exposure
    :param saturation_target_ppm: How many pixels that should be "saturated"
      (above ``target_value``), is parts per milion (1 pixel per megapixel)
    :param target_value: Target value. If ``None`` then the maximum value of
      the frame datatype ``-1`` is used.
    :param down_scale: The amount to down scale the frame in both height and
      width, this is done to speed up the calculations.
    :param max_allowed_exposure_time: The maximum allowed exposure time. Overrules the
      exposure time set by fps as long as its (<= max exposure time for the given fps)
    """

    def __init__(
        self,
        cam: qamlib.Camera,
        *,
        crop: tuple[int, int, int, int] | None = None,
        num_average: int = 3,
        # SimpleAutoExposure specific
        saturation_target_ppm: float = 100,
        target_value: int | None = None,
        down_scale: int = 4,
        max_allowed_exposure_time: int | None = None,
    ) -> None:
        self.__target_ppm = saturation_target_ppm

        self.target_value = target_value

        self._down_scale = down_scale

        self.max_allowed_exposure_time = max_allowed_exposure_time

        super().__init__(cam, crop=crop, num_average=num_average)

    def iterate(  # noqa C901
        self, meta: qamlib.FrameMetadata, frame: npt.NDArray[numpy.uint8]
    ) -> float | None:
        """Do one iteration of auto exposure.

        This function uses the metadata make sure that the frame is not older than
        the last time the exposure was changed.

        :param meta: The frame metadata.
        :param frame: The frame to look at.

        :raises ValueError: If the pixelformat is not supported (currently
            raw bayer formats are not supported).
        :raises ValueError: If the calculated exposure is above or below the
            boundary values.

        :returns: Calculated exposure ratio
        """
        if meta.time < self.last_exposure_change:
            log.debug("Got frame that is older than last exposure change. Skipping")
            return None

        self._n_frames += 1

        # Only use relevant elements for checking exposure level
        match self._format.pixelformat.fourcc:
            case "BGR3":
                pass
            case "BGR4":
                frame = frame[:, :, :3]
            case "HSV":
                frame = frame[:, :, :2]
            case "RGB3":
                pass
            case "RGB4":
                frame = frame[:, :, :3]
            case "GREY":
                pass
            case "GREY16":
                pass
            case _:
                raise ValueError(
                    "Auto exposure does not support current format "
                    f"({self._format.pixelformat.fourcc})"
                )

        if self._avg_frame is None:
            if self._crop is not None:
                self._avg_frame = frame[self._crop].astype(numpy.float64)
            else:
                self._avg_frame = frame.astype(numpy.float64)
        else:
            if self._crop is not None:
                self._avg_frame += frame[self._crop]
            else:
                self._avg_frame += frame

        if self._n_frames < self.num_average:
            return None

        self._avg_frame /= self._n_frames
        avg_frame = self._avg_frame

        width, height = avg_frame.shape[:2]
        avg_frame = cast(
            np.ndarray,
            cv2.resize(
                avg_frame,
                (int(height / self._down_scale), int(width / self._down_scale)),
            ),
        )

        # Calculate the target value
        max_val = numpy.iinfo(frame.dtype).max

        if self.target_value is None:
            target_val = max_val - 1
        else:
            if self.target_value >= max_val:
                raise ValueError(f"target_value is too high (>= {max_val})")
            elif self.target_value <= 0:
                raise ValueError("target_value is too low (<= 0)")

            target_val = self.target_value

        target_count = avg_frame.size * (self.__target_ppm / 1_000_000)

        histogram = numpy.bincount(
            avg_frame.astype(frame.dtype).ravel(), minlength=max_val + 1
        )

        exposure = self.cam.get_control("Exposure Time, Absolute")

        count = 0
        for i in range(max_val, target_val, -1):
            count += histogram[i]

        if count >= target_count:
            ratio = max(0.8, 1 - 0.02 * (count / target_count))
        else:
            highest_step_up = 5
            lowest_index = max_val // highest_step_up
            for i in range(target_val, lowest_index, -1):
                count += histogram[i]

                if count >= target_count:
                    ratio = target_val / i

                    # Limit potential overshoot be decreasing ratio a bit
                    ratio = max(1, ratio * 0.99)
                    break
            else:
                # Limit new exposure to 5x current exposure
                ratio = highest_step_up

        exposure = int(exposure * ratio)

        if self.max_allowed_exposure_time is not None:
            if self.max_allowed_exposure_time < self._exposure.min:
                raise ValueError(
                    f"max_allowed_exposure_time: Exceeded min={self._exposure.min} or"
                    f"max={self._exposure.max} limit. Please change the value."
                )
            elif exposure > self.max_allowed_exposure_time:
                exposure = self.max_allowed_exposure_time

        if exposure >= self._exposure.max:
            log.debug("Reached maximum exposure value")
        elif exposure <= self._exposure.min:
            log.debug("Reached minimum exposure value")

        self.last_exposure_change = time.clock_gettime(meta.clock)
        self.cam.set_control("Exposure Time, Absolute", exposure)

        self._avg_frame = None
        self._n_frames = 0

        return ratio


def one_shot_auto_exposure(
    cam: qamlib.Camera,
    error_percentage: float = 0.3,
    max_iterations: int = 200,
    auto_exposure: SimpleAutoExposure | None = None,
) -> None:
    """One time auto exposure.

    This function will do a round of auto exposure and return either when the
    targeted ``precision`` is reached, if ``max_iterations`` is reached, or if
    the exposure value reaches either the minimum or the maximum value.

    :param cam: The :class:`qamlib.Camera` to do auto exposure on.
    :param error_percentage: The value of change in exposure below which is accepted as
      the auto exposure being finished.
    :param max_iterations: The maximum number of frames to use for auto exposure,
      meaning that the auto exposure is stopped when we reach this limit.
    :param auto_exposure: :class:`AutoExposure` object to use, if ``None`` it
      defaults to :class:`SimpleAutoExposure`

    :raises ValueError: If ``error_percentage`` is not between 0 and 100.
    """
    if 0.0 >= error_percentage or error_percentage >= 100.0:
        raise ValueError("precision must be between 0 and 100")

    error = error_percentage / 100

    def iteration(auto_exposure: SimpleAutoExposure) -> None:
        for i in range(max_iterations):
            meta, frame = cam.get_frame()

            res = auto_exposure.iterate(meta, frame)

            exposure_value = cam.get_control("exposure time, absolute")

            if res is not None:
                if exposure.min >= exposure_value >= exposure.max:
                    log.debug(
                        f"Auto-exposure done after {i} frames. Exposure limit reached"
                    )
                    return

                diff = abs(res - 1)
                if diff <= error:
                    log.debug(f"Auto-exposure done after {i} frames.")
                    return
        log.debug("Auto-exposure done. Max iterations reached")

    exposure = cam.list_controls()["exposure_time_absolute"]

    if auto_exposure is None:
        auto_exposure = SimpleAutoExposure(cam)

    # TODO When qamlib has a way to check if it is already streaming, then only
    # start the stream if needed.
    with cam:
        iteration(auto_exposure)


if __name__ == "__main__":
    cam = qamlib.Camera()
    logging.basicConfig(level=logging.DEBUG)
    one_shot_auto_exposure(cam)
