"""Functions for doing camera calibration."""

import logging

import numpy as np
import numpy.typing as npt
import qamlib


log = logging.getLogger("quarklib")


def _get_average_frame(cam: qamlib.Camera, count: int = 5) -> npt.NDArray[np.float64]:
    """Get the average of ``count`` frames."""
    _, frame = cam.get_frame()
    avg: npt.NDArray[np.float64] = frame.astype(np.float64)
    for _ in range(count - 1):
        _, frame = cam.get_frame()
        avg += frame

    avg_frame = avg / count
    return avg_frame


def _adjust_down(
    cam: qamlib.Camera,
    control: qamlib.Control,
    target_value: float,
    threshold: float,
) -> None:
    """Adjust a control downwards towards a target value."""
    with cam:
        if np.average(_get_average_frame(cam)) > 40:
            log.warning("Frame is quite saturated, make sure the sensor is covered")

        cur_val = cam.get_control(control.id)
        max_iterations = ((control.max - control.min) // control.step) + 1

        log.debug("Adjusting black level up until we reach the threshould")
        for _ in range(max_iterations):
            avg_frame = _get_average_frame(cam)
            log.debug(f"Got average value: {np.average(avg_frame):.4f}")

            count = np.count_nonzero(avg_frame <= target_value)
            if count / avg_frame.size <= threshold:
                break

            cur_val = cam.get_control(control.id)

            if cur_val == control.max:
                log.warning(f"Got to maximum value for {control.name}")
                return

            log.debug(f"Setting black_level: {cur_val + control.step}")
            cam.set_control(control.id, cur_val + control.step)

            # Throw away a few frames to make sure were getting frames with the
            # new black level
            _get_average_frame(cam)
        else:
            log.warning("Reached maximum iterations ({max_iterations})")

        log.debug(
            "Adjusting black level down until we reach the threshould and are done"
        )
        for _ in range(max_iterations):
            cur_val = cam.get_control(control.id)

            avg_frame = _get_average_frame(cam)
            log.debug(f"Got average value: {np.average(avg_frame):.4f}")

            count = np.count_nonzero(avg_frame <= target_value)
            frac = count / avg_frame.size
            log.debug(f"Got {frac * 100:2f}% of pixels below threshold")
            if frac > threshold:
                log.debug(f"Final black_level: {cur_val}")
                break

            if cur_val == control.min:
                log.warning(f"Got to minimum value for {control.name}")
                return

            log.debug(f"Setting black_level: {cur_val - control.step}")
            cam.set_control(control.id, cur_val - control.step)

            # Throw away a few frames to make sure were getting frames with the
            # new black level
            _get_average_frame(cam)
        else:
            log.warning("Reached maximum iterations ({max_iterations})")


def calibrate_black_level(
    cam: qamlib.Camera,
    *,
    target_value: float | None = None,
    threshold: float = 0.01,
) -> None:
    """Calibrate Black Level for a camera.

    This function will average 5 frames and count the number of pixels that are
    below or equal to ``target_value``, and adjust the black level until there
    are more than ``threshold`` specifies at that target.

    :param cam: The :class:`qamlib.Camera` object to do the calibration to.
    :param target_value: The target brightness value, if not set, it is chosen
      according to the sensor type.
    :param threshold: The fraction of pixels that should be ``<= target_value``
      before the calibration stops.

    :raises ValueError:
        - If using a camera with an unsupported sensor type. Currently, `IMX`,
            `GMAX`, and `CMV` are supported.
        - If the camera does not support a greyscale format.
    """
    sensor = cam.get_ext_control("sensor_type").value

    if target_value is None:
        if "IMX" in sensor:
            target_value = 2
        elif "GMAX" in sensor:
            target_value = 2
        elif "CMV" in sensor:
            target_value = 1
        else:
            raise ValueError(f"Sensor type ({sensor}) not supported")

    ctrls = cam.list_controls()

    fmts = cam.list_formats()
    fmt = "GREY"
    if fmt not in fmts:
        raise ValueError("No grey format found")

    old_fmt = cam.get_format()
    cam.set_format(fmt)

    # Set to 8-bits
    old_bitmode = cam.get_control("sensor_bit_mode")
    cam.set_control("sensor_bit_mode", 0)

    # Set low exposure time
    old_exp = cam.get_control("exposure time, absolute")
    cam.set_control("exposure time, absolute", 100)

    # Set high FPS for faster calibration
    old_fps = cam.get_framerate()
    cam.set_framerate(90)

    try:
        # Start at the default value for the black level
        cam.set_control("black_level", ctrls["black_level"].default_value)
        _adjust_down(cam, ctrls["black_level"], target_value, threshold=threshold)
    finally:
        # Restore previous settings
        cam.set_format(old_fmt)
        cam.set_framerate(old_fps)
        cam.set_control("sensor_bit_mode", old_bitmode)
        cam.set_control("exposure_time_absolute", old_exp)
