"""Qtec ISP related functions."""

import logging

import qamlib


log = logging.getLogger("quarklib")


def create_trigger_sequence(
    cam: qamlib.Camera,
    exposures: list[int],
    *,
    flash_times: list[int] | None = None,
    set_control: bool = True,
) -> None:
    """Create a trigger sequence with minimal frame delay.

    :param cam: The :class:`qamlib.Camera` class for which the trigger
      sequence is for.
    :param exposures: A sequential list of the exposure times for the trigger
      sequence.
    :param flash_times: Optional flash_times for each exposure.
    :param set_control: Whether to actually set the "Trigger Sequence" control
      for ``cam``.

    :raises ValueError:
        - If the camera does not support setting the trigger sequence.
        - If `flash_times` is defined and is not the same length as `exposures`.
    """
    if "trigger_sequence" not in cam.list_controls():
        raise ValueError(
            "No trigger sequence control present, check that it is supported "
            "for this sensor and bitstream"
        )

    trig_seq = qamlib.TriggerSequenceValue()

    # Used to calculate minimal frame delay
    fot = cam.get_control("frame overhead time")
    rot = cam.get_control("read-out time")

    if flash_times is not None and len(exposures) != len(flash_times):
        raise ValueError(
            f"flash_times ({len(flash_times)}) should be of same "
            f"length as exposures ({len(exposures)})"
        )

    # Create trigger sequence with minimal delay
    for i in range(len(exposures)):
        exp = exposures[i]
        next_i = (i + 1) % len(exposures)

        if flash_times is None:
            delay = max(exp + fot, exp + fot + rot - exposures[next_i])
        else:
            delay = max(exp + fot, exp + fot + rot - exposures[next_i], flash_times[i])

        trig_seq.add_exposure(exp, exp, delay, 0)

    if set_control:
        cam.set_ext_control("trigger sequence", trig_seq)

    return trig_seq
