"""Automatic White Balance methods.

This module provides algorithms for performing automatic white balance correction.
Currently, only the Grey World method is implemented.

Example usage using the iterative helper function:
>>> cam = qamlib.CameraController()
>>> print(  # doctest: +ELLIPSIS
...     one_shot_auto_white_balance(
...         cam=cam,
...         num_average=5,
...         max_iterations=10,
...         target_error=0.01,
...         pixel_value_interval=(30, 230),
...     )
... )
[..., ..., ...]
"""

import logging
import warnings
from abc import ABC
from abc import abstractmethod
from typing import Optional

import numpy as np
import numpy.typing as npt
import qamlib

from quarklib.camera import Camera  # type: ignore


log = logging.getLogger("quarklib")


GAIN_INT_FRACTION = 16384


def get_average_frame(cam: qamlib.Camera, n: int):
    """Capture multiple frames and return their average.

    :param cam: The camera to use.
    :param n: How many frames to capture.
    :returns: The average frame.
    """
    _, frame = cam.get_frame()
    for _ in range(n - 1):
        frame = frame.astype(np.float64)
        _, new_frame = cam.get_frame()
        frame += new_frame

    return frame / n


def _check_frame(frame, fail_ratio, pixel_value_interval):
    flattened = _threshold_saturation(
        frame,
        pixel_value_interval,
    )
    selected_ratio = len(flattened) / (frame.shape[0] * frame.shape[1])
    if selected_ratio < fail_ratio:
        raise RuntimeError(
            "Image is either too dark or too saturated. "
            f"Measured ratio is {selected_ratio}, minimum ratio is {fail_ratio}"
        )

    return flattened.mean(axis=0)


class AutoWhiteBalance(ABC):
    """Base class for automatic white balance methods.

    White balance requires the pixel format to be in either RGB(A) or BGR(A)
    formats.

    :param cam: Camera used for auto white balance
    :param crop: Optional frame crop

    :raises ValueError:
        - If the pixel format is invalid.
        - If the camera does not support digital gain mapping.
    """

    def __init__(
        self,
        cam: Camera,
        crop: tuple[int, int, int, int] | None = None,
    ):
        self.cam = cam
        digital_gain = cam.controls.get("digital_gain")

        if digital_gain is None:
            raise ValueError("Camera does not support setting digital gain.")

        self.digital_gain = digital_gain

        pixel_format = self.cam.get_format().pixelformat.fourcc
        if pixel_format in ["BGR3", "BGR4"]:
            self._is_bgr = True
        elif pixel_format in ["RGB3", "RGB4"]:
            self._is_bgr = False
        else:
            raise ValueError(
                f"Invalid pixel format {pixel_format!r} for camera. "
                "Must be either 'RGB' or 'BGR'"
            )

        if pixel_format in ["BGR4", "RGB4"]:
            self._has_alpha = True
        else:
            self._has_alpha = False

        self.gains = np.array([1.0, 1.0, 1.0])

        if crop is not None:
            s1 = slice(crop[1], crop[1] + crop[3])  # Width crop
            s2 = slice(crop[0], crop[0] + crop[2])  # Height crop
            self._crop: tuple[slice, slice] | None = (s1, s2)
        else:
            self._crop = None

    def _get_channels(self, values: np.ndarray) -> np.ndarray:
        if self._has_alpha:
            if self._is_bgr:
                return values[:, :, :3]
            else:
                return values[:, :, 1:]
        else:
            return values

    def _set_gain(self, values: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """Set camera digital gain values.

        The function converts from floating point to the internal 16-bit integer value
        used on the camera.

        :param values: List of values.
        :returns: The rounded written values (the precision is 1/(2**16)).
        """
        if self._is_bgr:
            values = np.array([values[2], values[1], values[0]])

        self.digital_gain.write(values)
        ret_vals = self.digital_gain.read()

        if self._is_bgr:
            return np.array([ret_vals[2], ret_vals[1], ret_vals[0]])
        else:
            return ret_vals

    @property
    def gains(self) -> npt.NDArray[np.float32]:
        """Get normalized camera gain values as a list.

        :returns: List of normalized gain values in range [0, 4]
        """
        return self._gains

    @gains.setter
    def gains(self, values: npt.NDArray[np.float32]):
        actual_values = self._set_gain(values)
        self._gains = actual_values

    def _crop_frame(self, frame: np.ndarray) -> np.ndarray:
        if self._crop:
            return frame[self._crop]
        else:
            return frame

    @abstractmethod
    def iterate(
        self,
        frame: np.ndarray,
        *,
        fail_ratio: float = 0.3,
        pixel_value_interval: tuple[int, int] = (40, 220),
    ) -> float:
        """Perform an iteration of the white balancing algorithm.

        The method returns the angular error of the white balance of the frame prior to
        correction. See :func:`angular_error` for details on the calculation.

        :param frame: The input frame
        :param fail_ratio: If the total fraction of pixels removed due to either low
            pixel values or high saturation exceeds this value, the method fails and
            raises a RuntimeError.
        :param pixel_value_interval: Only pixels with average values within this
            interval are included in the calculation.
        :raises RuntimeError: An error is raised either due to over saturation (see
            `saturation_fail_ratio`) or if the target angular error is not reached
            within the maximum allowed number of iterations.
        :returns: The angular error of the mean pixel prior to correction.
        """
        ...

    def evaluate(self, frame: np.ndarray) -> tuple[float, float]:
        """Evaluate the current white balance.

        The method calculates the angular error for the provided frame.

        :param frame: Input image.
        :returns: The saturation error as a fraction ([0, 1]) and in degrees.
        """
        cropped = self._crop_frame(frame)
        cropped = self._get_channels(cropped)
        mean = cropped.mean(axis=(0, 1))
        return angular_error(mean)


class GreyWorld(AutoWhiteBalance):
    """Estimate gain values using the Grey World assumption.

    The algorithm uses a fixed color channel to determine the target luminosity for
    the other channels. The channel can either be set manually or be automatically
    determined during initialization.

    :param cam: Camera used for auto white balance
    :param crop: Optional frame crop
    :param target_value: Desired target value. By default, a number of
        sample frames are captured and the brightest mean channel value is used as
        the target. Note that this resets the current gain values.
    """

    def __init__(
        self,
        cam: qamlib.Camera,
        crop: tuple[int, int, int, int] | None = None,
        *,
        target_value: int | None = None,
    ):
        super().__init__(cam, crop=crop)
        self.target_value = target_value
        self.target_channel_idx = None

    def iterate(
        self,
        frame: np.ndarray,
        *,
        fail_ratio: float = 0.3,
        pixel_value_interval: tuple[int, int] = (40, 220),
    ) -> float:
        """Perform an iteration of the grey world white balancing algorithm.

        The method updates the gains using the following equation:
            gains_new = target_value * (gains_previous / mean_pixel_value),

        where mean_pixel_value is the average pixel value over the frame (or crop if
        specified), target_value is the target value for all color channels, and
        gains_previous are the current gain values for each channel.

        The method returns the angular error of the white balance of the frame prior to
        correction. See :func:`angular_error` for details on the calculation.

        :param frame: The input frame
        :param fail_ratio: If the total fraction of pixels removed due to either low
            pixel values or high saturation exceeds this value, the method fails and
            raises a RuntimeError.
        :param pixel_value_interval: Only pixels with average values within this
            interval are included in the calculation.

        :raises RuntimeError: An error is raised either due to over saturation (see
            `saturation_fail_ratio`) or if the target angular error is not reached
            within the maximum allowed number of iterations.

        :returns: The angular error of the mean pixel prior to correction.
        """
        self._validate_parameters(fail_ratio, pixel_value_interval)

        frame = self._crop_frame(frame)
        frame = self._get_channels(frame)
        mean_pixel_value = _check_frame(frame, fail_ratio, pixel_value_interval)

        # Initialize the target channel index if necessary
        if self.target_value is None and self.target_channel_idx is None:
            self.target_channel_idx = mean_pixel_value.argmax()

        if self.target_channel_idx is None:
            target_value = self.target_value
        else:
            target_value = mean_pixel_value[self.target_channel_idx]
        gains_new = target_value * (self.gains / mean_pixel_value)

        if np.any(np.array(gains_new) > 4.0):
            warnings.warn(
                "One or more color gains above the maximum of 4x. "
                f"Values: {gains_new} "
                "This will lead to inaccurate results",
                stacklevel=2,
            )
            gains_new = np.clip(
                gains_new,
                0,
                self.digital_gain.get("red").formatter.valid_set.upper  # type: ignore[union-attr]
                - self.digital_gain.get("red").formatter.precision,  # type: ignore[union-attr]
            )
        if np.any(np.array(gains_new) < 1.0 - 1e-4):
            warnings.warn(
                f"One or more color gains under 1x. Values: {gains_new} "
                "This will result in distorted colors for saturated pixels",
                stacklevel=2,
            )

        self.gains = gains_new
        logging.debug(f"Calculated gains for iteration: {gains_new}")

        error, _ = angular_error(mean_pixel_value)
        logging.debug(f"Saturation error for iteration: {error}")
        return error

    def _validate_parameters(self, fail_ratio, pixel_value_interval):
        if fail_ratio < 0 or fail_ratio >= 1:
            raise ValueError(
                f"Fail ratio {fail_ratio!r} is invalid"
                " It must be in the interval [0, 1)."
            )
        elif pixel_value_interval[0] < 0 or pixel_value_interval[1] > 255:
            raise ValueError(
                f"Pixel value interval {pixel_value_interval} is invalid."
                "Both values must be in the interval [0, 255]."
            )
        elif pixel_value_interval[0] > pixel_value_interval[1]:
            raise ValueError(
                "pixel_value_interval[0] is larger than pixel_value_interval[1]."
            )
        elif self.target_value is not None and (
            self.target_value > pixel_value_interval[1]
            or self.target_value < pixel_value_interval[0]
        ):
            raise ValueError(
                "Target pixel value is outside the allowable range",
                pixel_value_interval,
            )


def one_shot_auto_white_balance(
    cam: Optional[qamlib.Camera] = None,
    white_balance_method: AutoWhiteBalance | None = None,
    target_error: float = 0.01,
    max_iterations: int = 10,
    *,
    num_average: int = 10,
    fail_ratio: float = 0.3,
    pixel_value_interval: tuple[int, int] = (40, 220),
):
    """Perform estimation.

    Estimates gains iteratively using the mean pixel value a number of averaged
    frames. The method uses the brightest color channel as the target value. The
    gains are then estimated linearly from the observed pixel values in each
    iteration.

    Image saturation is estimated at the start to determine whether the image is
    suitable for determining white balance. If not, an error is raised. Each
    iteration averages several frames to improve stability. Over saturated pixels
    are removed before calculating the mean pixel value. The saturation value is
    calculated using the :func:`linear_saturation` function.

    As soon as the method reaches the target value, it returns. See
    :func:`angular_error` for details on the calculation.

    :param cam: Camera to use. Of omitted, the camera from the white_balance_method is
        used.
    :param white_balance_method: The white balance algorithm to use. If omitted, a
        GreyWorld method is created using the provided camera.
    :param target_error: Target saturation error [0, 1].
    :param max_iterations: Maximum number of iterations
    :param num_average: The number of frames to average in each
        iteration
    :param fail_ratio: If the total fraction of pixels with saturation
        above the threshold exceeds this value, the method fails and a RuntimeError
        is raised
    :param pixel_value_interval: Only pixels with average values within this interval
            are included in the calculation.

    :raises ValueError: When either `cam` or `white_balance_method` is `None`.
    :raises RuntimeError: An error is raised either due to over saturation (see
        `saturation_fail_ratio`) or if the target angular error is not reached
        within the maximum allowed number of iterations.

    :returns: The angular error
    """
    if cam is None and white_balance_method is None:
        raise ValueError("Must provide either `cam` or `white_balance_method`")

    if cam is None:
        cam = white_balance_method.cam  # type: ignore

    with cam:  # type: ignore
        if white_balance_method is None:
            white_balance_method = GreyWorld(cam)

        for _ in range(max_iterations):
            frame = get_average_frame(cam, num_average)
            error = white_balance_method.iterate(
                frame, fail_ratio=fail_ratio, pixel_value_interval=pixel_value_interval
            )
            if error < target_error:
                logging.debug("Successfully completed white balance calibration")
                return white_balance_method.gains
        else:
            raise RuntimeError("Could not achieve target angular error")


def angular_error(pixel: np.ndarray) -> tuple[float, float]:
    """Calculate the saturation of a pixel in RGB or BGR space.

    The saturation is calculated as the angle between the measured pixel and the grey
    vector ([1., 1., 1.]). The error is returned as a value in the interval [0, 1]
    where 1 is the worst-case error (~54.4 degrees). The absolute value, in degrees, is
    also returned.

    :param pixel: The measured pixel value (usually a mean)

    :raises ValueError: If the pixel value is 0.

    :returns: The error fraction and the absolute error in degrees.
    """
    if np.all(pixel == 0):
        raise ValueError("Invalid pixel value [0, 0, 0]. Pixels must be non-zero.")
    dot_pd = min(1, _unit_vector(pixel).dot(GREY_VEC))
    angle_radians = np.arccos(dot_pd)
    fraction = angle_radians / WORST_ERROR
    angle_degrees = angle_radians / np.pi * 180

    return fraction, angle_degrees


def minimum_value_for_error(e: float) -> int:
    """Calculate the minimum pixel value necessary for the given error.

    Because the pixel values are discrete, the minimum achievable error depends on the
    distance between the individual pixels. Since the angular error is used, the angle
    between pixels depend on their overall luminance, i.e., adjacent pixel values
    have a larger angular deviation for low luminance values.

    It is therefore necessary to calculate a minimum pixel value to consider, otherwise
    the target error might not be achievable due to the inevitable rounding errors.

    The value is calculated by finding the intensity value *i* for which a pixel
    diverging from grey by exactly one value in one channel (e.g. [i, i, i+1]) has an
    angular error equal to e. In other words, the function solves the following equation
    for e:

    e = acos(v*g/(norm(v)norm(g))),

    where v is the diverging pixel value and g is the corresponding grey value
    ([i, i, i]).

    Example
    -------
        >>> print(minimum_value_for_error(e=0.01))
        49.01053561714026

    :param e: Target saturation error in the interval [0, 1].
    :returns: The minimum pixel channel value to achieve the target error.

    """
    e_rad = e * WORST_ERROR
    cos_b = np.cos(e_rad)
    det_sq = 6 * cos_b * np.sqrt(2) * np.sqrt(1 - cos_b**2)
    upper = -6 * cos_b**2 + 6 - det_sq
    ta = 18 * (cos_b**2 - 1)

    return upper / ta


def _unit_vector(vec: np.ndarray) -> np.ndarray:
    return vec / np.linalg.norm(vec)


GREY_VEC = _unit_vector(np.ones(3))
WORST_ERROR = np.arccos(GREY_VEC.dot([1, 0, 0]))


def _threshold_saturation(
    img: np.ndarray,
    pixel_value_interval,
):
    mean = np.mean(img, axis=2)
    return img[(mean > pixel_value_interval[0]) & (mean < pixel_value_interval[1])]
