# flake8-in-file-ignores: noqa: DOC201
"""LUT generation functions.

This module contains functions for generating some common look up tables
(LUTs). They assume that all values in ``0, 1, ..., 2^bits - 1`` are valid,
with ``bits`` being the chosen bit depth (default is 12).

They also support chaining by using the ``base`` argument, for some functions
chaining might not make that much sense but it is still possible (eg. linear
lut will not be linear if the base is not linear).
"""

import enum

import numpy as np
import numpy.typing as npt


def __init_res(base: np.ndarray | None, bits: int) -> np.ndarray:
    """Initialize the output array and check if it is valid."""
    max_val = 2**bits - 1
    num_elems = max_val + 1

    if base is None:
        # We have an entry for each value
        return np.arange(num_elems)
    else:
        if len(base) != num_elems:
            raise Exception(f"Base array is too long ({len(base)}")
        if np.max(base) > max_val:
            raise Exception(f"Base array has values that are too big (>{max_val})")
        if np.min(base) < 0:
            raise Exception("Base array has values that are too small (<0)")
        return base


def gamma(
    value: float,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create a LUT for gamma correction.

    :param value: The gamma value.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError: If the generated LUT contains values too large to
        represent in a `bits`-bit integer.
    """
    inv = 1.0 / value

    max_val = 2**bits - 1

    res = __init_res(base, bits)

    # Calculate LUT
    res = (res / max_val) ** inv * max_val

    if np.max(res) > max_val:
        raise ValueError("Got invalid LUT value >MAX")

    return res.astype(dtype)


def intensity_stretch(
    min_value: int,
    max_value: int,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create quantization LUT.

    This LUT will stretch the values between ``min_value`` and ``max_value`` to
    the full value range ``0`` to ``2^bits-1``.

    :param min_value: Minimum value.
    :param max_value: Maximum value.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError:
        - If `min_value` is not between 0 and `max_value`.
        - If `max_value` is not between 0 and the maximum value for an
            `bits`-bit integer.
    """
    max_val = 2**bits - 1

    if not 0 <= min_value <= max_value:
        raise ValueError("min_value must be between 0 and max_value")
    if not 0 <= max_value <= max_val:
        raise ValueError(f"max_value must be between 0 and maximum value ({max_val})")

    res = __init_res(base, bits)

    scale = max_val / (max_value - min_value)
    res = (res - min_value) * scale
    res = np.clip(res, 0, max_val)

    return res.astype(dtype)


def intensity_stretch_percentage(
    min_percentage: float,
    max_percentage: float,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create quantization LUT.

    This LUT will stretch the values between ``min_percentage`` and
    ``max_percentage`` to the full value range ``0`` to ``2^bits-1``.

    :param min_percentage: Minimum value [%].
    :param max_percentage: Maximum value [%].
    :param base: Base LUT, if ``None`` the identity LUT is used as base.

    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.

    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError:
        - If `min_percentage` is not between 0 and `max_percentage`.
        - If `max_percentage` is not between 0 and 100.
    """
    if not 0 <= min_percentage <= max_percentage:
        raise ValueError("min_percentage must be between 0 and max_percentage")
    if not 0 <= max_percentage <= 100:
        raise ValueError("min_percentage must be between 0 and 100")

    max_val = 2**bits - 1

    min_value = round(max_val * min_percentage / 100)
    max_value = round(max_val * max_percentage / 100)

    return intensity_stretch(min_value, max_value, base=base, bits=bits, dtype=dtype)


def invert(
    *, base: np.ndarray | None = None, bits: int = 12, dtype: npt.DTypeLike = np.int32
) -> np.ndarray:
    """Create inverted LUT.

    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.
    """
    max_val = 2**bits - 1

    res = __init_res(base, bits)

    res = max_val - res

    return res.astype(dtype)


def linear(
    degree: float,
    y_intersection: int,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create linear LUT.

    :param degree: The slope in degrees.
    :param y_intersection: The value at which the y-axis is intersected (x=0).
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.
    """
    max_val = 2**bits - 1

    res = __init_res(base, bits)

    res = res * np.tan(np.radians(degree)) + y_intersection

    res = np.clip(res, 0, max_val)

    return res.astype(dtype)


def linear_percentage(
    degree: float,
    y_intersection_percentage: int,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create linear LUT.

    :param degree: The slope in degrees.
    :param y_intersection_percentage: The percentage at which the y-axis is
      intersected (x=0).
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.
    """
    max_val = 2**bits - 1

    intersection = max_val * (y_intersection_percentage / 100)

    return linear(degree, intersection, base=base, bits=bits, dtype=dtype)


def piecewise_linear(
    points: list[tuple[int, int]],
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create LUT with multiple linear pieces.

    This generates a LUT that is linear between the given ``points``.

    :param points: A list of points to interpolate between (x, y). The points
      should be sorted in accending order by ``x`` value.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError:
        - If any coordinate `x` or `y` in `points` is not between 0 and 100.
        - If the `x` coordinates in `points` are not non-decreasing.
    """
    max_val = 2**bits - 1

    res = __init_res(base, bits)

    prev_x = 0
    prev_y = 0
    for x, y in points:
        if (not 0 <= x <= max_val) or (not 0 <= y <= max_val):
            raise ValueError(f"Point coordinates needs to be between 0 and {max_val}")
        if x <= prev_x:
            raise ValueError("The points x coordinate needs to increasing")
        diff = y - prev_y
        splice = res[prev_x:x]

        splice = splice - np.min(splice)
        splice = splice / np.max(splice) * diff
        splice += prev_y

        res[prev_x:x] = splice
        prev_x = x
        prev_y = y

    diff = max_val - prev_y
    splice = res[prev_x:]

    splice = splice - np.min(splice)
    splice = splice / np.max(splice) * diff
    splice += prev_y

    res[prev_x:] = splice

    return res.astype(dtype)


def piecewise_linear_percentage(
    points: list[tuple[int, int]],
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create LUT with multiple linear pieces.

    :param points: A list of points to interpolate between (x, y). The values
      should be as a percentage of the max value.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError: If any coordinate `x` or `y` in `points` is not between
        0 and 100.
    """
    max_val = 2**bits - 1

    value_coords = []
    for x, y in points:
        if (not 0 <= x <= 100) or (not 0 <= y <= 100):
            raise ValueError(
                "Point (percentage) coordinates needs to be between 0 and 100"
            )
        x = max_val * (x / 100)
        y = max_val * (y / 100)
        value_coords.append((x, y))

    return piecewise_linear(value_coords, base=base, bits=bits, dtype=dtype)


def quantization(
    steps: int,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create quantization LUT.

    :param steps: The number of steps.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError: If `steps` is not between 1 and the maximum value of a
        `bit`-bit integer.
    """
    max_val = 2**bits - 1

    if not 1 < steps <= max_val:
        raise ValueError(f"steps must be between 1 and maximum value ({max_val})")

    res = __init_res(base, bits)

    # Base mask
    mask = res < 0

    step_size = max_val / (steps - 1)
    for step in range(steps):
        step_limit = round((step + 1) * step_size)
        step_val = round(step * step_size)
        mask = res < step_limit
        mask = mask ^ (res < step_val)
        res[mask] = step_val

    return res.astype(dtype)


def range_value(
    low: int,
    high: int,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create range LUT.

    This generates a lut for which values below ``low`` and above ``high` are
    set to ``0`` and all other values are set to ``2^bits``

    :param low: The lower threshold.
    :param high: The upper threshold.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError:
        - If `low` is not between 0 and `high`.
        - If `high` is not in the valid range of a `bit`-bit integer.
    """
    max_val = 2**bits - 1

    if not 0 <= low <= high:
        raise ValueError("low must be between 0 and high")
    if not 0 <= high <= max_val:
        raise ValueError(f"high must be between 0 and max value ({max_val})")

    res = __init_res(base, bits)

    mask = res < low
    mask += res > high
    res[mask] = 0
    res[~mask] = max_val

    return res.astype(dtype)


def range_percentage(
    low: float,
    high: float,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create range LUT.

    This generates a lut for which values below ``low`` and above ``high` are
    set to ``0`` and all other values are set to ``2^bits``

    :param low: The lower threshold, as a percentage of the maximum value
      (``2^bits-1``).
    :param high: The upper threshold, as a percentage of the maximum value
      (``2^bits-1``).
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError:
        - If `low` is not between 0 and `high`.
        - If `high` is not between 0 and 100.
    """
    if not 0 <= low <= high:
        raise ValueError("low_percentage must be between 0 and high_percentage")
    if not 0 <= high <= 100:
        raise ValueError("high_percentage must be between 0 and 100")

    max_val = 2**bits - 1
    low = int(round(max_val / 100 * low))
    high = int(round(max_val / 100 * high))

    return range_value(low, high, base=base, bits=bits, dtype=dtype)


def s_curve(
    mid_point: int,
    force: float,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create S-curve LUT.

    :param mid_point: Where on the X axis the midpiont should be.
    :param force: The "bendiness" of the curve, as a percentage.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError:
        - If `mid_point` is not in the valid range of an `bit`-bit integer.
        - If `force` is not between 0 and 100.
    """
    max_val = 2**bits - 1

    if not 0 <= mid_point <= max_val:
        raise ValueError("low must be between 0 and high")

    if not 0 <= force <= 100:
        raise ValueError("force must be between 0 and 100")

    res = __init_res(base, bits)

    # This calculation is taken from gwt
    # https://gitlab.com/qtec/software/distro/camera_utilities/qteclib/-/blob/master/src/lut/lut_options.c
    # We disable lint for variables starting with uppercase.
    K = 1.0  # noqa N806
    M = max_val  # noqa N806
    N = float(mid_point)  # noqa N806
    Pmin = 1  # noqa N806

    # Avoid division by zero
    if K == N:
        K /= 10  # noqa N806

    Z = (((N * N - M * M) / (M - N)) * ((N * N - M * M) / (M - N))) / (  # noqa N806
        # noqa N806
        (K - M) * (K - N)
    )
    Pmax = (3 * K + Z * (K - M)) / (Z * (K - M) + 3 * M)  # noqa N806

    P = (force / 100) * (Pmax - Pmin) + Pmin  # noqa N806

    a = 1000 * (P - 1) / ((K - M) * (K - N))
    b = (a / 1000) * (N * N - M * M) / (M - N)
    c = 1 + (a / 1000) * M * N

    res = (a * (res**3)) / 1000 + b * (res**2) + c * res

    return res.astype(dtype)


def s_curve_percentage(
    mid_point: int,
    force: float,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create S-curve LUT.

    :param mid_point: Where on the X axis the midpiont should be as a percentage.
    :param force: The "bendiness" of the curve, as a percentage.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError: If `mid_point` is not between 0 and 100
    """
    if not 0 <= mid_point <= 100:
        raise ValueError("mid_point should be between 0 and 100")

    max_val = 2**bits - 1

    mid_point_value = mid_point / 100 * max_val

    return s_curve(mid_point_value, force, base=base, bits=bits, dtype=dtype)


class Threshold(enum.Enum):
    """Threshold types for a threshold LUT."""

    BINARY = enum.auto()
    TRUNCATE = enum.auto()
    TO_ZERO = enum.auto()
    TO_ZERO_INVERTED = enum.auto()


def threshold(
    value: int,
    type: Threshold,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create threshold LUT.

    :param value: The threshold value.
    :param type: The threshold type.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError: If `value` is not in the valid range for an `bits`-bit
        integer.
    """
    max_val = 2**bits - 1

    if not 0 <= value <= max_val:
        raise ValueError(f"value must be between 0 and maximum value ({max_val})")

    res = __init_res(base, bits)

    larger = res > value
    if type == Threshold.BINARY:
        res[larger] = max_val

        # '~' inverts the mast
        res[~larger] = 0
    elif type == Threshold.TRUNCATE:
        larger = res > value
        res[larger] = value
    elif type == Threshold.TO_ZERO:
        # '~' inverts the mast
        res[~larger] = 0
    elif type == Threshold.TO_ZERO_INVERTED:
        res[larger] = max_val

    return res.astype(dtype)


def threshold_percentage(
    percentage: float,
    type: Threshold,
    *,
    base: np.ndarray | None = None,
    bits: int = 12,
    dtype: npt.DTypeLike = np.int32,
) -> np.ndarray:
    """Create threshold LUT.

    :param percentage: The threshold percentage.
    :param type: The threshold type.
    :param base: Base LUT, if ``None`` the identity LUT is used as base.
    :param bits: The number of bits that the LUT should use, eg 8 bits results
      in a LUT of size 2^8.
    :param dtype: The NumPy dtype for the resulting LUT.

    :raises ValueError: If `percentage` is not between 0 and 100.
    """
    if not 0 <= percentage <= 100:
        raise ValueError("value must be between 0 and 100")

    max_val = 2**bits - 1
    value = round(max_val / 100.0 * percentage)

    return threshold(value, type, base=base, bits=bits, dtype=dtype)
